/*****************************************************************************
 *  rspwrmeter_callbacks.c
 *
 *  Rohde & Schwarz Power Meter Instrument Driver
 *  Original Release: August 2014
 *  By: Jiri Kominek, Lukas Hynek
 *
 *  Should you have any technical questions please contact the hotline of
 *  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
 *
 *  e-mail: CustomerSupport@rohde-schwarz.com
 *
 *  Version: 1.0.0
 *
 *  Copyright (c) 2014, Rohde & Schwarz GmbH & Co. KG
 *  All rights reserved.
 *****************************************************************************/

#pragma warning (disable:4100 4055)

#include "rspwrmeter.h"
#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__))
    #define ANSI
#endif
#ifndef ANSI                /* UNIX compatible */
    #include <sys/timeb.h>
    #include <sys/types.h>
    #include <unistd.h>
    #define Sleep sleep
#else                       /* ANSI compatible */
        #include <Windows.h>
#endif

/*****************************************************************************
 *   Read Callbacks ----------------------------------------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rspwrmeter_empty_ReadCallback
 * Purpose:  This function overrides standard callback function. Returns value
 *           from cache
 *****************************************************************************/
ViStatus rspwrmeter_empty_ReadCallback(ViSession          instrSession,
                                       ViConstString      repCapName,
                                       RsCoreAttributePtr attr)
{
    return VI_SUCCESS;
}
/*****************************************************************************
 * Function: rspwrmeter_ReadOnlyCollection_ReadCallback
 * Purpose:  This function overrides standard callback function. Returns value
 *           from cache
 *****************************************************************************/
ViStatus rspwrmeter_ReadOnlyCollection_ReadCallback(ViSession          instrSession,
                                                    ViConstString      repCapName,
                                                    RsCoreAttributePtr attr)
{
    return RsCore_ReadCallback(instrSession, repCapName, attr);
}

/*****************************************************************************
 *- Write Callbacks ---------------------------------------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rspwrmeter_empty_WriteCallback
 * Purpose:  This function overrides standard callback function. Does not write
 *           anything to instrument, just updates the cache
 *****************************************************************************/
ViStatus rspwrmeter_empty_WriteCallback(ViSession          instrSession,
                                        ViConstString      repCapName,
                                        RsCoreAttributePtr attr,
                                        void               *value)
{
    return VI_SUCCESS;
}

/*****************************************************************************
 *------------------------ Global Session Callbacks -------------------------*
 *****************************************************************************/

/*****************************************************************************
 * Function: rspwrmeter_logging_ReadCallback
 * Purpose:  This function overrides standard callback function. Returns value
 *           from cache
 *****************************************************************************/
ViStatus rspwrmeter_logging_ReadCallback(ViSession          instrSession,
                                         ViConstString      repCapName,
                                         RsCoreAttributePtr attr)
{
    return VI_SUCCESS;
}

/*****************************************************************************
 * Function: rspwrmeter_logging_WriteCallback
 * Purpose:  This function overrides standard callback function.
 *****************************************************************************/
ViStatus rspwrmeter_logging_WriteCallback(ViSession          instrSession,
                                          ViConstString      repCapName,
                                          RsCoreAttributePtr attr,
                                          void               *value)
{
    return VI_SUCCESS;
}

/*****************************************************************************
* Function: rspwrmeter_TriggerSource_ReadCallback
* Purpose:  Used for reading channel-specific trigger source attribute
*            Exception for the NRX, it uses two SCPI commands for trigger source: TRIG{0}:MODE and TRIG{0}:SOURCE
*            NRP2 and NRPxxSN use the standard read procedure
*****************************************************************************/
ViStatus rspwrmeter_TriggerSource_ReadCallback(ViSession          instrSession,
                                               ViConstString      repCapName,
                                               RsCoreAttributePtr attr)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar   response[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar   repCapChannelCmd[RS_MAX_SHORT_MESSAGE_BUF_SIZE];
    ViInt32  rtNumValue;

    // If no command available, exit the function
    if (attr->command == NULL)
        return VI_SUCCESS;

    if (attr->dataType != RS_VAL_INT32)
    {
        checkErr(RsCore_GenerateInvalidDataTypeError(instrSession, attr->dataType, "rspwrmeter_TriggerSource_ReadCallback only accepts type RS_VAL_INT32"));
    }

    if (!rspwrmeter_IsNrx(instrSession))
    {
        checkErr(RsCore_ReadCallback(instrSession, repCapName, attr));
    }
    else
    {
        checkErr(RsCore_GetRepCapCmdValue(instrSession, "Channel", repCapName, RS_MAX_SHORT_MESSAGE_BUF_SIZE, repCapChannelCmd, NULL));
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "TRIG%s:MODE?", repCapChannelCmd);
        checkErr(RsCore_QueryViString(instrSession, cmd, response));
        if (strcmp(response, "FRE") == 0)
        {
            checkErr(RsCore_FindRangeTableDiscreteNumericValue(instrSession, attr->rangeTableDiscrete, "IMM", &rtNumValue));
            RsCore_SetAttributeP2Value(instrSession, attr, &rtNumValue);
        }
        else
        {
            checkErr(RsCore_ReadCallback(instrSession, repCapName, attr));
        }
    }

Error:
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_TriggerSource_WriteCallback
* Purpose:  Used for writing channel-specific trigger source attribute
*            Exception for the NRX, it uses two SCPI commands to set the trigger
*            NRP2 and NRPxxSN use the standard write procedure
*****************************************************************************/
ViStatus rspwrmeter_TriggerSource_WriteCallback(ViSession          instrSession,
                                                ViConstString      repCapName,
                                                RsCoreAttributePtr attr,
                                                void               *value)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar    repCapChannelCmd[RS_MAX_SHORT_MESSAGE_BUF_SIZE];
    ViInt32   int32Value;
    ViInt32   rangeTableNumValue;
    ViChar    *p2cmdString;
    ViBoolean trigAuto;

    // If no command available, exit the function
    if (attr->command == NULL)
        return VI_SUCCESS;

    if (attr->dataType != RS_VAL_INT32)
    {
        checkErr(RsCore_GenerateInvalidDataTypeError(instrSession, attr->dataType, "rspwrmeter_TriggerSource_WriteCallback only accepts type RS_VAL_INT32"));
    }

    if (!rspwrmeter_IsNrx(instrSession))
    {
        checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, value));
    }
    else
    {
        int32Value = *(ViInt32 *)value;
        checkErr(RsCore_GetRepCapCmdValue(instrSession, "Channel", repCapName, RS_MAX_SHORT_MESSAGE_BUF_SIZE, repCapChannelCmd, NULL));
        // For NRX, Immediate trigger is TRIG1:MODE AUTO
        checkErr(RsCore_FindRangeTableDiscreteCmdString(instrSession, attr->rangeTableDiscrete, int32Value, &p2cmdString));
        if (strcmp(p2cmdString, "IMM") == 0)
        {
            // For Immediate trigger
            snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "TRIG%s:MODE FRE", repCapChannelCmd);
            checkErr(RsCore_Write(instrSession, cmd));

            // Now normal write callback with internal trigger source
            checkErr(RsCore_FindRangeTableDiscreteNumericValue(instrSession, attr->rangeTableDiscrete, "INT", &rangeTableNumValue));
            int32Value = (ViInt32)rangeTableNumValue;
            checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, &int32Value));
        }
        else
        {
            // For other triggers always send TRIG:MODE AUTO / NORM based on the RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO
            RsCore_GetAttributeViBoolean(instrSession, repCapName, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO, 0, &trigAuto);
            snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "TRIG%s:MODE %s", repCapChannelCmd, trigAuto == VI_TRUE? "AUTO" : "NORM");
            checkErr(RsCore_Write(instrSession, cmd));

            // Now normal write callback with the input value
            checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, value));
        }
    }

Error:
    return error;
}


/*****************************************************************************
* Function: rspwrmeter_TriggerSourceAll_WriteCallback
* Purpose:  Used for writing global trigger source attribute
*            Exception for the NRX, it uses two SCPI commands to set the trigger
*            NRP2 and NRPxxSN use the standard write procedure
*            Same as the rspwrmeter_TriggerSource_WriteCallback, but without Channel parameter
*****************************************************************************/
ViStatus rspwrmeter_TriggerSourceAll_WriteCallback(ViSession          instrSession,
                                                   ViConstString      repCapName,
                                                   RsCoreAttributePtr attr,
                                                   void               *value)
{
    ViStatus error = VI_SUCCESS;
    ViInt32  int32Value;
    ViInt32  rangeTableNumValue;
    ViChar   *p2cmdString;

    // If no command available, exit the function
    if (attr->command == NULL)
        return VI_SUCCESS;

    if (attr->dataType != RS_VAL_INT32)
    {
        checkErr(RsCore_GenerateInvalidDataTypeError(instrSession, attr->dataType, "rspwrmeter_TriggerSource_WriteCallback only accepts type RS_VAL_INT32"));
    }

    if (!rspwrmeter_IsNrx(instrSession))
    {
        checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, value));
    }
    else
    {
        int32Value = *(ViInt32 *)value;
        // For NRX, Immediate trigger is TRIG1:MODE AUTO
        checkErr(RsCore_FindRangeTableDiscreteCmdString(instrSession, attr->rangeTableDiscrete, int32Value, &p2cmdString));
        if (strcmp(p2cmdString, "IMM") == 0)
        {
            // For Immediate trigger
            checkErr(RsCore_Write(instrSession, "TRIG:ALL:MODE FRE"));

            // Now normal write callback with internal trigger source
            checkErr(RsCore_FindRangeTableDiscreteNumericValue(instrSession, attr->rangeTableDiscrete, "INT", &rangeTableNumValue));
            int32Value = (ViInt32)rangeTableNumValue;
            checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, &int32Value));
        }
        else
        {
            // for all other triggers
            checkErr(RsCore_Write(instrSession, "TRIG:ALL:MODE NORM"));

            // Now normal write callback with the input value
            checkErr(RsCore_WriteCallback(instrSession, repCapName, attr, value));
        }
    }

Error:
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_ReadSensorInfo_ReadCallback
*            Returns string including quotes or doublequotes
*            For NRX a different command is needed compared to the NRP2
*****************************************************************************/
ViStatus rspwrmeter_ReadSensorInfo_ReadCallback(ViSession          instrSession,
                                                ViConstString      repCapName,
                                                RsCoreAttributePtr attr)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar    response[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar    repCapChannelCmd[RS_MAX_SHORT_MESSAGE_BUF_SIZE];
    ViBoolean isBaseUnit;

    if (attr->dataType != RS_VAL_STRING)
    {
        checkErr(RsCore_GenerateInvalidDataTypeError(instrSession, attr->dataType, "rspwrmeter_ReadSensorInfo_ReadCallback only accepts type RS_VAL_STRING"));
    }

    checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
    if (!isBaseUnit)
    {
        // Single sensor
        checkErr(RsCore_QueryViString(instrSession, "SYST:INFO?", response));
        RsCore_TrimString(response, RS_VAL_TRIM_WHITESPACES_AND_ALL_QUOTES);
        RsCore_SetAttributeP2Value(instrSession, attr, response);
    }
    else if (rspwrmeter_IsNrx(instrSession))
    {
        // NRX
        checkErr(RsCore_GetRepCapCmdValue(instrSession, "Channel", repCapName, RS_MAX_SHORT_MESSAGE_BUF_SIZE, repCapChannelCmd, NULL));
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS%s:INF?", repCapChannelCmd);
        checkErr(RsCore_QueryViString(instrSession, cmd, response));
        RsCore_TrimString(response, RS_VAL_TRIM_WHITESPACES_AND_ALL_QUOTES);
        RsCore_SetAttributeP2Value(instrSession, attr, response);
    }
    else
    {
        //NRP2, Use the standard callback
        checkErr(RsCore_ReadCallback(instrSession, repCapName, attr));
    }

Error:
    return error;
}
