/*****************************************************************************
*  rspwrmeter_utility.c
*
*  Rohde & Schwarz Power Meter Instrument Driver
*  Original Release: August 2014
*  By: Jiri Kominek, Lukas Hynek
*
*  Should you have any technical questions please contact the hotline of
*  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
*
*  e-mail: CustomerSupport@rohde-schwarz.com
*
*  Version: 1.0.1
*
*  Copyright (c) 2014, Rohde & Schwarz GmbH & Co. KG
*  All rights reserved.
*****************************************************************************/
#include <float.h>
#include <math.h>
#include "rspwrmeter.h"


/*****************************************************************************
*- Callback Declarations ---------------------------------------------------*
*****************************************************************************/

/*****************************************************************************
* Function: rspwrmeter_InitAttributes
* Purpose:  This function initializes attributes to the desired values if needed.
*****************************************************************************/
ViStatus rspwrmeter_InitAttributes(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    // - Driver Identification
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_DESCRIPTION, 0, RSPWRMETER_SPECIFIC_DRIVER_DESCRIPTION));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_PREFIX, 0, RSPWRMETER_SPECIFIC_DRIVER_PREFIX));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_LOCATOR, 0, RSPWRMETER_SPECIFIC_DRIVER_LOCATOR));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_VENDOR, 0, RSPWRMETER_SPECIFIC_DRIVER_VENDOR));

    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MAJOR_VERSION, 0, RSPWRMETER_CLASS_SPEC_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_CLASS_SPEC_MINOR_VERSION, 0, RSPWRMETER_CLASS_SPEC_MINOR_VERSION));

    // - Driver Capabilities
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_SUPPORTED_INSTRUMENT_MODELS, 0, RSPWRMETER_SUPPORTED_INSTRUMENT_MODELS));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_GROUP_CAPABILITIES, 0, RSPWRMETER_GROUP_CAPABILITIES));
    checkErr(RsCore_SetAttributeViString(instrSession, "", RS_ATTR_FUNCTION_CAPABILITIES, 0, RSPWRMETER_FUNCTION_CAPABILITIES));

    // - Version Info
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MAJOR_VERSION, 0, RSPWRMETER_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_VERSION, 0, RSPWRMETER_MINOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_SPECIFIC_DRIVER_MINOR_MINOR_VERSION, 0, RSPWRMETER_MINOR_MINOR_VERSION));

    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_MAJOR_VERSION, 0, RSPWRMETER_CLASS_SPEC_MAJOR_VERSION));
    checkErr(RsCore_SetAttributeViInt32(instrSession, "", RS_ATTR_CLASS_DRIVER_MINOR_VERSION, 0, RSPWRMETER_CLASS_SPEC_MINOR_VERSION));

    checkErr(RsCore_SetSpecificDriverRevision(instrSession, RS_ATTR_SPECIFIC_DRIVER_REVISION));

Error:
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_DefaultInstrSetup
* Purpose:  This function sends a default setup to the instrument. The
*           rspwrmeter_reset function calls this function. The
*           rspwrmeter_RsInit function calls this function when the
*           user passes VI_FALSE for the reset parameter. This function is
*           useful for configuring settings that other instrument driver
*           functions require.
*
*           Note:  Call this function only when the session is locked.
*****************************************************************************/
ViStatus rspwrmeter_DefaultInstrSetup(ViSession instrSession)
{
    ViStatus         error = VI_SUCCESS;
    ViChar           cmd[RS_MAX_MESSAGE_LEN];
    int              highestChannel;
    ViBoolean        channel1, channel2, channel3, channel4;
    RsCoreSessionPtr rsSession = NULL;
    ViBoolean        isBaseUnit;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    // Set all the attributes to the default state. Do not update inherent attributes!
    checkErr(RsCore_ApplyAttributeDefaults(instrSession, VI_FALSE));

    // Init attributes
    checkErr(rspwrmeter_InitAttributes(instrSession));

    rsSession->fastSweepInstrument = 0;
    checkErr(RsCore_ResetRegistersEseSre(instrSession));

    if (!RsCore_Simulating(instrSession))
    {
        checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
        if (isBaseUnit)
        {
            if (rspwrmeter_IsNrx(instrSession))
            {
                // NRX Base Unit
                checkErr(RsCore_Write(instrSession, "FORM:BORD NORM;:FORM:SREG ASC"));

                // Assign sensors to NRX Measurements
                checkErr(rspwrmeter_GetBaseUnitSensorsPresence(instrSession, &channel1, &channel2, &channel3, &channel4));

                snprintf(cmd, RS_MAX_MESSAGE_LEN, "CALC1:SENS:IND %d", channel1 ? 1 : 0);
                checkErr(RsCore_Write(instrSession, cmd));

                snprintf(cmd, RS_MAX_MESSAGE_LEN, "CALC2:SENS:IND %d", channel2 ? 2 : 0);
                checkErr(RsCore_Write(instrSession, cmd));

                snprintf(cmd, RS_MAX_MESSAGE_LEN, "CALC3:SENS:IND %d", channel3 ? 3 : 0);
                checkErr(RsCore_Write(instrSession, cmd));

                snprintf(cmd, RS_MAX_MESSAGE_LEN, "CALC4:SENS:IND %d", channel4 ? 4 : 0);
                checkErr(RsCore_Write(instrSession, cmd));

                // NRX Window Layout
                highestChannel = 0;
                if (channel1) highestChannel = 1;
                if (channel2) highestChannel = 2;
                if (channel3) highestChannel = 3;
                if (channel4) highestChannel = 4;

                if (highestChannel > 0)
                {
                    snprintf(cmd, RS_MAX_MESSAGE_LEN, "DISP:LAY L%d", highestChannel);
                    checkErr(RsCore_Write(instrSession, cmd));
                }
            }
            else
            {
                // NRP2 Base Unit
                checkErr(RsCore_Write(instrSession, "FORM:BORD SWAP;:FORM:SREG ASC"));
            }
        }
        else
        {
            // single sensors - NRPxxSN, NRQ6, NRPM
            checkErr(RsCore_Write(instrSession, "FORM:BORD NORM;:FORM:SREG ASC"));

            if (RsCore_IsInstrumentModel(instrSession, "NRQ6"))
            {
                checkErr(RsCore_Write(instrSession, "SENS:TRAC:IQ:DATA:FORM IQP"));
            }

            if (rsSession->sessionType == RS_INTF_USB)
            {
                // Single USB Sensors have problems with tout when viReadSTB()
                rsSession->viReadStbVisaTimeoutMs = 50;
            }
        }
    }

Error:
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_CheckStatus
* Purpose:  This function checks the status of the instrument to detect
*           whether the instrument has encountered an error. This function
*           is called at the end of most exported driver functions. If the
*           instrument reports an error, this function returns
*           RS_ERROR_INSTRUMENT_SPECIFIC. The user can set the
*           rspwrmeter_ATTR_QUERY_INSTRUMENT_STATUS attribute to VI_FALSE to disable this
*           check and increase execution speed.
*
*           Note:  Call this function only when the session is locked.
*****************************************************************************/
ViStatus rspwrmeter_CheckStatus(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    if (RsCore_QueryInstrStatus(instrSession) && !RsCore_Simulating(instrSession))
    {
        checkErr(RsCore_CheckStatus(instrSession, VI_SUCCESS));
    }

Error:
    return error;
}

/****************************************************************************************************
 *    rspwrmeter_SetOPCTimeout
    Sets new OPC timeout
******************************************************************************************************/
ViStatus rspwrmeter_SetOPCTimeout(ViSession instrSession,
                                  ViInt32   timeout)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_SetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, timeout));

Error:
    return error;
}

/****************************************************************************************************
 *    rspwrmeter_GetOPCTimeout
    Returns current OPC timeout
******************************************************************************************************/
ViStatus rspwrmeter_GetOPCTimeout(ViSession instrSession,
                                  ViInt32   *timeout)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_GetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, timeout));

Error:
    return error;
}

/****************************************************************************************************
 *    rspwrmeter_IsNrx
    Returns True, if the current instrument is a NRX Base Unit
******************************************************************************************************/
ViBoolean rspwrmeter_IsNrx(ViSession instrSession)
{
    return RsCore_IsInstrumentModel(instrSession, "NRX");
}

/****************************************************************************************************
 *    rspwrmeter_ParseUserDataContent
    Parses one User Data Content block, returns recent data
******************************************************************************************************/
ViStatus rspwrmeter_ParseUserDataContent(ViSession instrSession,
                                         ViChar    *data,
                                         ViInt32   count,
                                         ViInt32   arraySize,
                                         ViReal64  outputData[],
                                         ViInt32   *resultType,
                                         ViInt32   *retValues,
                                         ViChar    **pRecentData,
                                         ViInt32   *recentCount)
{  
    ViStatus error = VI_SUCCESS;
    ViString typeArr[] = { "AVG","MIN","MAX","RND",NULL };
    ViString dataTypeArr[] = { "f",NULL };

    ViChar      *pbuffer = data;
    ViChar      buffer[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar      fmt[RS_MAX_MESSAGE_LEN];
    ViInt32     dataType;
    ViInt32     byteLen = 0;
    ViInt32     int01;
    ViInt32     i;
    ViInt32     dataSizeNumber;
    ViInt32     numValues;
    ViInt32     retVal = 0;
    ViInt32     dataLen = 0;
    instrSession;

    // Read type
    (void)sscanf(pbuffer, "%3s", buffer);
    int01 = RsCore_FindStringIndex(typeArr, buffer);
    if (int01 <0) {
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE)
    };
    if (resultType) *resultType = int01;
    pbuffer += strlen(buffer);

    // Read data type
    (void)sscanf(pbuffer, "%1s", buffer);
    dataType = RsCore_FindStringIndex(dataTypeArr, buffer);
    switch (dataType) {
    case 0:
        byteLen = 4;
        break;
    default:
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE);
    }
    // Read number of numbers for data size
    pbuffer += strlen(buffer);
    retVal = sscanf(pbuffer, "%1ld", &dataSizeNumber);
    if (retVal<0) {
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE)
    };
    pbuffer++;

    snprintf(fmt, RS_MAX_MESSAGE_LEN, "%%%ldd", dataSizeNumber);
    retVal = sscanf(pbuffer, fmt, &numValues);  // NOLINT(clang-diagnostic-format-nonliteral)
    if (retVal<0) {
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE)
    };
    pbuffer += dataSizeNumber;

    // Count data length
    dataLen = byteLen * numValues;

    if (dataLen > count) {
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE)
    };

    switch (dataType) {
    case 0:   // Real32, 4 bytes
    {
        ViReal32    *temp = (ViReal32 *)pbuffer;

        byteLen = sizeof(temp);
        for (i = 0; i < numValues; i++)
        {
            if (i<arraySize)
                outputData[i] = (ViReal64)*temp++;
            pbuffer += byteLen;
        }
    }
    break;
    default:
        checkErr(RSPWRMETER_ERROR_INCORRECT_RESPONSE)
    }

    if (retValues) *retValues = numValues;
    if (recentCount) *recentCount = (ViInt32)(count - (pbuffer - data));
    if (pRecentData) *pRecentData = pbuffer;

Error:
    return error;
}

ViStatus rspwrmeter_utility_error_message(ViSession instrSession,
                                          ViStatus  errorCode,
                                          ViChar    errorMessage[])
{
   ViStatus error = VI_SUCCESS;

   (void)(RsCore_LockSession(instrSession));

   static RsCoreStringValueTable errorTable =
   {
       RSPWRMETER_ERROR_CODES_AND_MSGS,
       {0, NULL}
   };

   // all VISA and RS error codes are handled as well as codes in the table
   viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

   checkErr(RsCore_GetSpecificDriverStatusDesc(instrSession, errorCode, errorMessage, 256, errorTable));

Error:
   (void)RsCore_UnlockSession(instrSession);  // IGNORE: ERROR!!! Missing Lock
   return error;
}