/*****************************************************************************
*  rspwrmeter.c
*
*  Rohde & Schwarz Power Meter Instrument Driver
*  Original Release: August 2014
*  By: Jiri Kominek, Lukas Hynek
*
*  Should you have any technical questions please contact the hotline of
*  Rohde & Schwarz Vertriebs-GmbH Rohde & Schwarz Support Center
*
*  e-mail: CustomerSupport@rohde-schwarz.com
*
*  Version: 1.7.0
*
*  Modification History:
*    see ChangeLog
*
*  Copyright (c) 2014, Rohde & Schwarz GmbH & Co. KG
*  All rights reserved.
*****************************************************************************/
#include "rspwrmeter.h"


/*****************************************************************************
*- Value Definition and Range Tables ---------------------------------------*
*****************************************************************************/
static ViConstString  gsc_BooleanArr[] = {"OFF", "ON",NULL};
static ViConstString gsc_StatusRegArr[] = {
    "OPER", "QUES", "DEV", "OPER:CAL", "OPER:LLF", "OPER:ULF", "OPER:MEAS", "OPER:SENS", "OPER:TRIG", "QUES:CAL",
    "QUES:POW", "QUES:WIND", NULL
};
static ViConstString gsc_RegTypeArr[] = {"COND", "EVEN", NULL};


/*****************************************************************************
*- Callback Declarations ---------------------------------------------------*
*****************************************************************************/

/*****************************************************************************
*- User-Callable Functions (Exportable Functions) --------------------------*
*****************************************************************************/

/// HIFN  This function performs instrument driver initialization actions
/// HIRET Returns the status code of this operation.
/// HIPAR resourceName/Pass the resource name of the device to initialize.
/// HIPAR IDQuery/Specify whether you want the instrument driver to perform an ID Query.
/// HIPAR resetDevice/Specify whether you want the to reset the instrument during the
/// HIPAR resetDevice/initialization procedure.
/// HIPAR newInstrSession/Returns a ViSession handle that you use to identify the instrument in
/// HIPAR newInstrSession/all subsequent instrument driver function calls.
ViStatus _VI_FUNC rspwrmeter_init(ViRsrc resourceName, ViBoolean IDQuery,
                                  ViBoolean resetDevice, ViSession* newInstrSession)
{
    ViStatus error = VI_SUCCESS;

    if (newInstrSession == NULL)
    {
        (void)RsCore_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER,
                                  VI_ERROR_PARAMETER4, "Null address for Instrument Handle");

        checkErr(RS_ERROR_INVALID_PARAMETER);
    }

    checkErr(rspwrmeter_InitWithOptions(resourceName, IDQuery, resetDevice, "", newInstrSession));

Error:
    return error;
}

/// HIFN  This function performs instrument driver initialization actions
/// HIRET Returns the status code of this operation.
/// HIPAR resourceName/Pass the resource name of the device to initialize.
/// HIPAR IDQuery/Specify whether you want the instrument driver to perform an ID Query.
/// HIPAR resetDevice/Specify whether you want the to reset the instrument during the
/// HIPAR resetDevice/initialization procedure.
/// HIPAR optionString/You can use this control to set the initial value of certain
/// HIPAR optionString/attributes for the session.  The following table lists the attributes
/// HIPAR optionString/and the name you use in this parameter to identify the attribute.
/// HIPAR newInstrSession/Returns a ViSession handle that you use to identify the instrument in
/// HIPAR newInstrSession/all subsequent instrument driver function calls.
ViStatus _VI_FUNC rspwrmeter_InitWithOptions(
    ViRsrc resourceName,
    ViBoolean IDQuery,
    ViBoolean resetDevice,
    ViString optionString,
    ViSession* newInstrSession
)
{
    ViStatus error = VI_SUCCESS;
    RsCoreAttributePtr* attrList = NULL;
    ViSession instrSession = 0;
    RsCoreSessionPtr rsSession = NULL;

    if (newInstrSession == NULL)
    {
        (void)RsCore_SetErrorInfo(0, VI_FALSE, RS_ERROR_INVALID_PARAMETER, VI_ERROR_PARAMETER5,
                                  "Null address for Instrument Handle");
        checkErr(RS_ERROR_INVALID_PARAMETER);
    }

    attrList = (RsCoreAttributePtr*)rspwrmeter_RsCoreAttributeList;

    checkErr(RsCore_NewSpecificDriver(
        resourceName, // VISA Resource Name
        "rspwrmeter", // driver prefix
        optionString, // Init Options string, is applied to the session settings
        attrList, // List of all attributes from the rsxxx_AttrPropertiesList
        rspwrmeter_RsCoreAttributeListLength, // Attributes count
        VI_TRUE, //idnModelFullName Model has the full *IDN? query name e.g. true: "NRP33SN", VI_FALSE: "NRP" Used in CheckInstrument()
        0, // WriteDelay
        0, // ReadDelay
        1000000, // IO Segment Size
        RS_VAL_OPCWAIT_STBPOLLING, // OPC Wait Mode
        RSPWRMETER_OPC_TIMEOUT, // OPC timeout
        10000, // VISA Timeout
        600000, // Self-test timeout
        RS_VAL_BIN_FLOAT_FORMAT_SINGLE_4BYTES, // BinaryFloatNumbersFormat
        RS_VAL_BIN_INTEGER_FORMAT_INT32_4BYTES, // binaryIntegerNumbersFormat
        &instrSession));

    // Minimal allowed rscore version is 4.7.0
    checkErr(RsCore_CheckMinimalEngineVersion(instrSession, 4, 7, 0));

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));

    // No SCPI command has been sent yet.
    // Call viClear before sending any SCPI command
    checkErr(RsCore_ViClear(instrSession));

    /* --- Here perform settings that are default for this driver,
    but can be overwritten by the optionsString settings */

    rsSession->autoSystErrQuery = VI_FALSE;

    // Parse option string and optionally sets the initial state of the following session attributes
    checkErr(RsCore_ApplyOptionString(instrSession, optionString));
    checkErr(RsCore_BuildRepCapTable(instrSession, rspwrmeter_RsCoreRepCapTable, rspwrmeter_RsCoreRepCapTableCount));

    // Query *IDN? string, parse it and set the following attributes:
    // -RS_ATTR_INSTRUMENT_MANUFACTURER
    // - RS_ATTR_INSTRUMENT_MODEL - based on the idnModelFullName true: "NRP33SN", false : "NRP"
    // - RS_ATTR_INSTRUMENT_FIRMWARE_REVISION
    checkErr(RsCore_QueryAndParseIDNstring(instrSession, RSPWRMETER_SIMULATION_ID_QUERY, NULL));

    // Query *OPT? string, parse the options, remove the duplicates, sort them and store the result string to RS_ATTR_OPTIONS_LIST
    checkErr(RsCore_QueryAndParseOPTstring(instrSession, RSPWRMETER_SIMULATION_OPT_QUERY, RS_VAL_OPTIONS_PARSING_AUTO));

    // Default Instrument Setup + optional *RST
    if (resetDevice == VI_TRUE)
    {
        checkErr(rspwrmeter_reset(instrSession));
    }
    else
    {
        checkErr(rspwrmeter_DefaultInstrSetup(instrSession));
    }

    if (IDQuery == VI_TRUE)
    {
        checkErr(RsCore_FitsIDNpattern(instrSession, RSPWRMETER_VALID_ID_RESPONSE_STRING, NULL));
    }

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    RsCore_AfterErrorInInit(instrSession, rsSession, error, newInstrSession);
    return error;
}

/// HIFN  Performs zeroing using the signal at the sensor input(s). The
/// HIFN  sensor(s) must be disconnected from all power sources.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_ZeroAll(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_CHANNEL_ZERO_ALL_ONCE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Perform fast zeroing. However, the commands cannot be called unless
/// HIFN  the sensor(s) is(are) in the Trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_FastZeroAll(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_CHANNEL_FAST_ZERO_ALL_ONCE, NULL));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method queries return the value ON if a calibration is in
/// HIFN  progress, otherwise value OFF.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR zeroingComplete/This control returns the value ON if a calibration is in progress,
/// HIPAR zeroingComplete/otherwise value OFF.
ViStatus _VI_FUNC rspwrmeter_IsZeroCompleteAll(ViSession  instrSession,
                                               ViBoolean* zeroingComplete)
{
    ViStatus error = VI_SUCCESS;
    ViInt32  idx;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_QueryStringIndex(instrSession, "CAL:ALL:ZERO:AUTO?", gsc_BooleanArr, &idx));
    *zeroingComplete = (ViBoolean)idx;

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the sensor to the selected measurement mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR measurementMode/This control specifies the measurement mode to which sensor should be
/// HIPAR measurementMode/switched.
ViStatus _VI_FUNC rspwrmeter_MeasurementMode(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   measurementMode)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MODE, measurementMode),
            3, "Measurement Mode");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Activates the measurement of additional measured values that are
/// HIFN  determined together with the main measured value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR auxiliary/Selects the auxiliary mode.
ViStatus _VI_FUNC rspwrmeter_SetAuxiliary(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   auxiliary)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AUXILIARY,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AUXILIARY, auxiliary),
            3, "Auxiliary");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function queries which measurement results are available in the
/// HIFN  Trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR auxiliary/Returns the auxiliary mode.
ViStatus _VI_FUNC rspwrmeter_GetAuxiliary(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32*  auxiliary)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AUXILIARY,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AUXILIARY, auxiliary),
            3, "Auxiliary");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects the output unit for the measured power values.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR units/Selects the output unit for the measured power values.
ViStatus _VI_FUNC rspwrmeter_ConfigureUnits(ViSession instrSession,
                                            ViInt32   calculateBlock,
                                            ViInt32   units)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_BLOCK_UNITS,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_MEAS_BLOCK_UNITS, units),
            3, "Units");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Turns the "Auto" range function on/off.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR rangeAutoEnabled/Turns the "Auto" range function on/off.
ViStatus _VI_FUNC rspwrmeter_ConfigureRangeAutoEnabled(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean rangeAutoEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_RANGE_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_RANGE_AUTO, rangeAutoEnabled),
            3, "Range Auto Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects a measurement range in which the corresponding sensor is to
/// HIFN  perform a measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR path/Selects a measurement range in which the corresponding sensor is to
/// HIPAR path/perform a measurement.
ViStatus _VI_FUNC rspwrmeter_ConfigureRangePath(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   path)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_RANGE_PATH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_RANGE_PATH, path),
            3, "Path");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the cross-over level. Shifts the transition ranges between the
/// HIFN  measurement ranges. This may improve the measurement accuracy for
/// HIFN  special signals, i.e. signals with a high crest factor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR crossoverLevel/Sets the cross-over level.
ViStatus _VI_FUNC rspwrmeter_ConfigureRangeCrossoverLevel(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  crossoverLevel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_RANGE_CROSSOVER_LEVEL,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_RANGE_CROSSOVER_LEVEL, crossoverLevel),
            3, "Crossover Level");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Toggles between automatic and manual filter setting.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR averagingAutoEnabled/Toggles between automatic and manual filter setting.
ViStatus _VI_FUNC rspwrmeter_ConfigureAveragingAutoEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean averagingAutoEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_COUNT_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_COUNT_AUTO, averagingAutoEnabled),
            3, "Averaging Auto Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Informs the instrument about the frequency of the power to be measured
/// HIFN  since this frequency is not automatically determined. The frequency is
/// HIFN  used to determine a frequency-dependent correction factor for the
/// HIFN  measurement results.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR correctionFrequency/The frequency of the input signal.
ViStatus _VI_FUNC rspwrmeter_ConfigureCorrectionFrequency(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  correctionFrequency)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_FREQUENCY,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_FREQUENCY, correctionFrequency),
            3, "Correction Frequency");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures display settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR displayState/Opens or closes a window.
/// HIPAR displayName/Determines the string to be displayed in the titlebar of the
/// HIPAR displayName/respective window.
/// HIPAR displayResolution/Specifies the number of significant places of the mantissa when the
/// HIPAR displayResolution/measurement result is displayed on a linear scale.
/// HIPAR timeslot/Sets the displayed timeslot, if the primary or secondary channel is
/// HIPAR timeslot/operated in the Timeslot mode.
/// HIPAR displayMode/Selects the display mode for measured data.
/// HIPAR additionalInfo/Determines which additional information about the measured values is
/// HIPAR additionalInfo/shown in the display (right-hand side, next to the main measured
/// HIPAR additionalInfo/values).
/// HIPAR errorlistState/Displays ON or removes OFF a windows which contains the contents of
/// HIPAR errorlistState/the SCPI error queue.
ViStatus _VI_FUNC rspwrmeter_ConfigureDisplay(ViSession instrSession,
                                              ViInt32   channel,
                                              ViBoolean displayState,
                                              ViString  displayName,
                                              ViInt32   displayResolution,
                                              ViInt32   timeslot,
                                              ViInt32   displayMode,
                                              ViInt32   additionalInfo,
                                              ViBoolean errorlistState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_DISPLAY_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_ENABLED, displayState),
            3, "Display State");
    if (displayState)
    {
        viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_NAME, displayName),
                4, "Display Name");

        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_RESOLUTION, displayResolution),
                5, "Display Resolution");

        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_TIMESLOT, timeslot),
                6, "Timeslot");

        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_MODE, displayMode),
                7, "Display Mode");

        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_ADDITIONAL_INFO, additionalInfo),
                8, "Additional Info");

        viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_DISPLAY_ERRORLIST_ENABLED, errorlistState),
                9, "Error List State");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method automatically determines scaling for the analog display.
/// HIFN  The upper and the lower limit value of the display is set as a function
/// HIFN  of the current measurement data.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
ViStatus _VI_FUNC rspwrmeter_AutoScalingAnalogMeter(ViSession instrSession,
                                                    ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "DISP:WIND%ld:MET:AUTO ONCE", channel);
    else
        strcpy(cmd, "DISP:MET:AUTO ONCE");
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the lower and upper limit value of the analog scale.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR left/Sets the lower limit value of the analog scale.
/// HIPAR right/Sets the upper limit value of the analog scale.
ViStatus _VI_FUNC rspwrmeter_ConfigureAnalogMeterScaling(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViReal64  left,
                                                         ViReal64  right)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER, left),
            3, "Left");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_UPPER, right),
            4, "Right");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the lower and upper power limit.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR lowerPower/Sets the lower power.
/// HIPAR upperPower/Sets the upper power.
ViStatus _VI_FUNC rspwrmeter_ConfigureAnalogMeterPowerRange(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViReal64  lowerPower,
                                                            ViReal64  upperPower)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER_POWER,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER_POWER, lowerPower),
            3, "Lower Power");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_UPPER_POWER, upperPower),
            4, "Upper Power");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets a lower and upper power ratio limit.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR lowerPowerRatio/Sets the lower ratio.
/// HIPAR upperPowerRatio/Sets the upper ratio.
ViStatus _VI_FUNC rspwrmeter_ConfigureAnalogMeterPowerRatioRange(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  lowerPowerRatio,
                            ViReal64  upperPowerRatio)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER_RATIO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_LOWER_RATIO, lowerPowerRatio),
            3, "Lower Power Ratio");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ANALOG_METER_UPPER_RATIO, upperPowerRatio),
            4, "Upper Power Ratio");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Effective for pulse analysis measurements. Enables or
///HINF  disables the display of the droop.
///HIRET Returns the status code of this operation.
///HIPAR channel/Numerical selection of a channel. Use 0 to omitt the
///HIPAR channel/channel numeric suffix.
///HIPAR droopEnabled/Effective for pulse analysis measurements.
///HIPAR droopEnabled/Enables or disables the display of the droop.
ViStatus _VI_FUNC rspwrmeter_ConfigureDisplayPowerPulseDroopEnabled(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean droopEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   channelRepCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_DISPLAY_POWER_PULSE_DROOP_ENABLED,
        "Channel", channel, RS_REPCAP_BUF_SIZE, channelRepCap), 2, "Channel");

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, channelRepCap, RSPWRMETER_ATTR_DISPLAY_POWER_PULSE_DROOP_ENABLED, droopEnabled),
            3, "Droop Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the video bandwidth for the Trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bandwidth/Sets the video bandwidth.
ViStatus _VI_FUNC rspwrmeter_SetVideoBandwidth(ViSession instrSession,
                                               ViInt32   channel,
                                               ViString  bandwidth)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_VBW,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_VBW, bandwidth),
            3, "Bandwidth");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Gets the video bandwidth for the Trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR bufferSize/This control defines the size of buffer passed to Bandwidth argument.
/// HIPAR bandwidth/This control returns video bandwidth.
ViStatus _VI_FUNC rspwrmeter_GetVideoBandwidth(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   bufferSize,
                                               ViChar    bandwidth[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_VBW,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_VBW,
            bufferSize, bandwidth),
            4, "Bandwidth");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Provides a list of strings with permissible settings for Video
/// HIFN  Bandwidth attribute.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferSize/This control defines the size of buffer passed to Bandwidth List
/// HIPAR bufferSize/argument.
/// HIPAR bandwidthList/This control returns the comma separated list of possible video
/// HIPAR bandwidthList/bandwidths.
ViStatus _VI_FUNC rspwrmeter_GetVideoBandwidthList(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   bufferSize,
                                                   ViChar    bandwidthList[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_VBW_LIST,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_VBW_LIST,
            bufferSize, bandwidthList),
            4, "Bandwidth List");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Resets the limit monitoring state and the internal counter for limit
/// HIFN  violations.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
ViStatus _VI_FUNC rspwrmeter_ResetRange(ViSession instrSession,
                                        ViInt32   calculateBlock)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, calculateBlock, 1, 8),
            2, "Calculate Block");

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC%ld:LIM:CLE", calculateBlock);
    checkErr(RsCore_Write(instrSession, cmd));
    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC%ld:LIM:CLE:AUTO ONCE", calculateBlock);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method stores the maximum and minimum values for each calculate
/// HIFN  block. The current measured value is stored as the new minimum and
/// HIFN  maximum value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
ViStatus _VI_FUNC rspwrmeter_ResetAuxVal(ViSession instrSession,
                                         ViInt32   calculateBlock)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, calculateBlock, 1, 8),
            2, "Calculate Block");

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC%ld:EXTR:RES", calculateBlock);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the monitoring function.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR monitoringState/Switches the monitoring function for the upper limit on or off.
/// HIPAR autoReset/Automatically resets the limit monitoring state and the internal
/// HIPAR autoReset/counter for limit violations.
ViStatus _VI_FUNC rspwrmeter_ConfigureMonitoringFunction(ViSession instrSession,
                                                         ViInt32   calculateBlock,
                                                         ViBoolean monitoringState,
                                                         ViBoolean autoReset)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_LIMITS_ENABLED,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_ENABLED, monitoringState),
            3, "Monitoring State");

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_AUTO_RESET, autoReset),
            4, "Auto Reset");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the lower and upper range values for a given
/// HIFN  calculate block.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR lowerEnabled/Switches the monitoring function for theupper and lower limit on or
/// HIPAR lowerEnabled/off.
/// HIPAR rangeLower/Sets a lower limit for the measured values.
/// HIPAR upperEnabled/Switches the monitoring function for the upper limit on or off.
/// HIPAR rangeUpper/Sets an upper limit for the measured values.
/// HIPAR warningBeep/Switches the acoustic warning for limit violations on or off.
ViStatus _VI_FUNC rspwrmeter_ConfigureRange(ViSession instrSession,
                                            ViInt32   calculateBlock,
                                            ViBoolean lowerEnabled,
                                            ViReal64  rangeLower,
                                            ViBoolean upperEnabled,
                                            ViReal64  rangeUpper,
                                            ViBoolean warningBeep)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_LIMITS_LOWER_ENABLED,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_LOWER_ENABLED, lowerEnabled),
            3, "Lower Enabled");
    if (lowerEnabled)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_LOWER, rangeLower),
                4, "Range Lower");
    }
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_UPPER_ENABLED, upperEnabled),
            5, "Upper Enabled");
    if (upperEnabled)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_UPPER, rangeUpper),
                6, "Range Upper");
    }

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_WARNING_BEEP, warningBeep),
            7, "Warning Beep");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets a lower and upper limit for the measured power.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR powerRangeLower/Sets a lower limit for the measured power.
/// HIPAR powerRangeUpper/Sets a upper limit for the measured power.
ViStatus _VI_FUNC rspwrmeter_ConfigurePowerRange(ViSession instrSession,
                                                 ViInt32   calculateBlock,
                                                 ViReal64  powerRangeLower,
                                                 ViReal64  powerRangeUpper)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_LIMITS_LOWER_POWER,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_LOWER_POWER, powerRangeLower),
            3, "Power Range Lower");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_UPPER_POWER, powerRangeUpper),
            4, "Power Range Upper");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets a lower and upper limit for the measured power ratios.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR powerRatioRangeLower/Sets a lower limit for the measured power ratios.
/// HIPAR powerRatioRangeUpper/Sets an upper limit for the measured power ratios.
ViStatus _VI_FUNC rspwrmeter_ConfigurePowerRatioRange(ViSession instrSession,
                                                      ViInt32   calculateBlock,
                                                      ViReal64  powerRatioRangeLower,
                                                      ViReal64  powerRatioRangeUpper)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_LIMITS_LOWER_RATIO,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_LOWER_RATIO, powerRatioRangeLower),
            3, "Power Ratio Range Lower");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_UPPER_RATIO, powerRatioRangeUpper),
            4, "Power Ratio Range Upper");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries whether upper or lower limits have been exceeded and  the
/// HIFN  number of limit violations that occurred.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR failed/Queries whether upper or lower limits have been exceeded.
/// HIPAR failCount/Queries the number of limit violations that occurred.
ViStatus _VI_FUNC rspwrmeter_QueryFailedRange(ViSession  instrSession,
                                              ViInt32    calculateBlock,
                                              ViBoolean* failed,
                                              ViInt32*   failCount)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_LIMITS_FAIL,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_FAIL, failed),
            3, "Failed");

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_LIMITS_FAIL_COUNT, failCount),
            4, "Fail Count");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Performs triggering and makes sure that the sensor directly changes
/// HIFN  from the WAIT_FOR_TRG state to the MEASURING state irrespective of the
/// HIFN  selected trigger source.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_ImmediateTrigger(ViSession instrSession,
                                              ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_IMMEDIATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_IMMEDIATE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the trigger signal source, slope, level, dropout, holdoff and
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR triggerSource/Sets the trigger signal source.
/// HIPAR triggerSlope/This attribute determines whether the rising (POSitive) or the falling
/// HIPAR triggerSlope/(NEGative) edge of the signal is used for triggering.
/// HIPAR level/Determines the power a trigger signal must exceed (TRIG:SLOP POS) or
/// HIPAR level/fall short of (TRIG:SLOP NEG) before a trigger event is detected. This
/// HIPAR level/setting is only used for the trigger signal source (TRIG:SOUR
/// HIPAR level/INTernal).
/// HIPAR dropout/Sets the dropout time. The dropout time prevents the trigger system
/// HIPAR dropout/from being activated by signal excursions prior to the trigger event
/// HIPAR dropout/targeted. The parameter is set to a value that is slightly higher than
/// HIPAR dropout/the maximum duration of power drops that are supposed to execute
/// HIPAR dropout/triggering.
/// HIPAR holdoff/Defines a period after a trigger event within which all further
/// HIPAR holdoff/trigger events are ignored.
/// HIPAR hysteresis/Sets the trigger hysteresis. The trigger hysteresis prevents the
/// HIPAR hysteresis/trigger system from being activated too early if the trigger threshold
/// HIPAR hysteresis/is slightly underranged or exceeded. Before triggering can occur again,
/// HIPAR hysteresis/the signal level must be lower (higher) than the trigger level minus
/// HIPAR hysteresis/(plus) hysteresis.
ViStatus _VI_FUNC rspwrmeter_ConfigureTrigger(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   triggerSource,
                                              ViInt32   triggerSlope,
                                              ViReal64  level,
                                              ViReal64  dropout,
                                              ViReal64  holdoff,
                                              ViReal64  hysteresis)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SOURCE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SOURCE, triggerSource),
            3, "Trigger Source");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SLOPE, triggerSlope),
            4, "Trigger Slope");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_LEVEL, level),
            5, "Trigger Level");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_DROPOUT, dropout),
            6, "Dropout");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_HOLDOFF, holdoff),
            7, "Holdoff");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_HYSTERESIS, hysteresis),
            8, "Hysteresis");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures trigger delay.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR delayAuto/Makes certain that a measurement is not started until the sensor has
/// HIPAR delayAuto/settled. To this end, the delay value is automatically determined. This
/// HIPAR delayAuto/is especially important when thermal sensors are used.
/// HIPAR delay/Defines the delay between the trigger event and the beginning of the
/// HIPAR delay/actual measurement (integration).
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerDelay(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean delayAuto,
                                                   ViReal64  delay)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_DELAY_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_DELAY_AUTO, delayAuto),
            3, "Delay Auto");
    if (!delayAuto)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_DELAY, delay),
                4, "Delay");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the selected sensor in trigger master state. If the trigger
/// HIFN  master detects an(internal) trigger event, it generates an external
/// HIFN  (hardware) trigger signal, which allows the synchronous triggering of
/// HIFN  other sensors.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR triggerMasterState/Sets the selected sensor in trigger master state. If the trigger
/// HIPAR triggerMasterState/master detects an(internal) trigger event, it generates an external
/// HIPAR triggerMasterState/(hardware) trigger signal, which allows the synchronous triggering of
/// HIPAR triggerMasterState/other sensors.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerMasterState(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean triggerMasterState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_MASTER_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_MASTER_ENABLED, triggerMasterState),
            3, "Trigger Master State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Enables or disables the trigger sender mode of the sensor. In this
/// HIFN  state, the power sensor can output a digital trigger signal in sync
/// HIFN  with its own trigger event. If enabled, select the output port for the
/// HIFN  trigger signal using rspwrmeter_ConfigureTriggerSenderPort.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omitt the channel numeric
/// HIPAR channel/suffix.
/// HIPAR triggerSenderState/Enables or disables the trigger sender mode of the sensor.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerSenderState(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViBoolean triggerSenderState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SENDER_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SENDER_ENABLED, triggerSenderState),
            3, "Trigger Sender State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the trigger master port.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR masterPort/Configures the trigger master port.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerMasterPort(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   masterPort)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_MASTER_PORT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_MASTER_PORT, masterPort),
            3, "Master Port");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects the port where the sensor outputs its own trigger event in
/// HIFN  case it is trigger sender.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omitt the channel numeric
/// HIPAR channel/suffix.
/// HIPAR senderPort/Selects the port where the sensor outputs its own trigger event in
/// HIPAR senderPort/case it is trigger sender.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerSenderPort(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   senderPort)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SENDER_PORT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SENDER_PORT, senderPort),
            3, "Sender Port");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  If set to ON, the command blocks the external trigger bus as long as
/// HIFN  the sensor is not in WAIT_FOR_TRIGGER state. It is usually used in
/// HIFN  combination with property Master Enabled and makes certain, that a new
/// HIFN  measurement will only be started after all sensors have completed their
/// HIFN  last measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR triggerSynchronizeState/If set to ON, the command blocks the external trigger bus as long as
/// HIPAR triggerSynchronizeState/the sensor is not in WAIT_FOR_TRIGGER state. It is usually used in
/// HIPAR triggerSynchronizeState/combination with property Master Enabled and makes certain, that a new
/// HIPAR triggerSynchronizeState/measurement will only be started after all sensors have completed their
/// HIPAR triggerSynchronizeState/last measurement.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerSynchronizeState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViBoolean triggerSynchronizeState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SYNCHRONIZE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_SYNCHRONIZE, triggerSynchronizeState),
            3, "Trigger Synchronize State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Selects the external connection for the sensor's sync output.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR triggerSynchronizePort/Selects the external connection for the sensor's sync output.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerSynchronizePort(ViSession instrSession,
                                                             ViInt32   triggerSynchronizePort)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_CHANNEL_TRIGGER_SYNCHRONIZE_PORT, triggerSynchronizePort),
            2, "Channel Trigger Synchronize Port");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  When ATrigger is set to ON, the WAIT_FOR_TRG state is automatically
/// HIFN  exited when no trigger event occurs within a period that corresponds to
/// HIFN  the reciprocal of the display update rate.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR autoMode/When ATrigger is set to ON, the WAIT_FOR_TRG state is automatically
/// HIPAR autoMode/exited when no trigger event occurs within a period that corresponds to
/// HIPAR autoMode/the reciprocal of the display update rate.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerAuto(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViBoolean autoMode)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO, autoMode),
            3, "Auto Mode");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the auto trigger execution state.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR executed/Returns the auto trigger execution state.
ViStatus _VI_FUNC rspwrmeter_GetTriggerAutoExecuted(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32*  executed)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO_EXECUTED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO_EXECUTED, executed),
            3, "Executed");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the state of the auto trigger level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR autoLevel/Configures the state of the auto trigger level.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerAutoLevel(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean autoLevel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO_LEVEL,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_AUTO_LEVEL, autoLevel),
            3, "Auto Level");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Configures the state of the auto trigger level.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR autoTriggerDelay/If Auto Trigger is enabled, sets the delay between the artificial trigger event and the beginning of the actual measurement.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerAutoDelay(ViSession instrSession,
                                                       ViReal64  autoTriggerDelay)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, "", RSPWRMETER_ATTR_CHANNEL_AUTO_TRIGGER_DELAY, autoTriggerDelay),
            2, "Auto Trigger Delay");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the external trigger impedance
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR impedance/Configures the external trigger impedance
ViStatus _VI_FUNC rspwrmeter_ConfigureExternalTriggerImpedance(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   impedance)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_EXTERNAL_IMPEDANCE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_EXTERNAL_IMPEDANCE, impedance),
            3, "Impedance");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Sets the number of measurement cycles to be performed when the measurement is started with Initiate.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR channelTriggerCount/Sets the number of measurement cycles to be performed when the measurement is started with Initiate.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerCount(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViInt32   channelTriggerCount)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRIGGER_COUNT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRIGGER_COUNT, channelTriggerCount),
            3, "Channel Trigger Count");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Available in trace mode. Defines the method how to cope with the
/// HIFN  misalignment between the trigger event and the sample point.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR jitterSuppressionMethod/Available in trace mode. Defines the method how to cope with the
/// HIPAR jitterSuppressionMethod/misalignment between the trigger event and the sample point.
ViStatus _VI_FUNC rspwrmeter_ConfigureTriggerJitterSuppressionMethod(
                            ViSession instrSession,
                            ViInt32   jitterSuppressionMethod)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, NULL, RSPWRMETER_ATTR_TRIGGER_JITTER_SUPPRESSION_METHOD, jitterSuppressionMethod),
            2, "Jitter Suppression Method");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}


/// HIFN  Available in trace mode. Supported for bandwidths  80 MHz.
/// HIFN     Queries the currently used value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR suppressionCurrentValue/Available in trace mode. Supported for bandwidths  80 MHz.
ViStatus _VI_FUNC rspwrmeter_GetTriggerJitterSuppressionCurrentValue(
                            ViSession instrSession,
                            ViInt32*  suppressionCurrentValue)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, NULL, RSPWRMETER_ATTR_TRIGGER_JITTER_SUPPRESSION_CURRENT_VALUE, suppressionCurrentValue),
            2, "Suppression Current Value");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}


/// HIFN  Available in trace mode.
/// HIFN     Queries the known trigger offset.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR suppressionTriggerOffset/Available in trace mode.
ViStatus _VI_FUNC rspwrmeter_GetTriggerJitterSuppressionTriggerOffset(
                            ViSession instrSession,
                            ViReal64* suppressionTriggerOffset)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, NULL, RSPWRMETER_ATTR_TRIGGER_JITTER_SUPPRESSION_TRIGGER_OFFSET, suppressionTriggerOffset),
            2, "Trigger Jitter Suppression Trigger Offset");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Performs a power down reset of the selected sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_SensorReboot(ViSession instrSession,
                                          ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_SENSOR_REBOOT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_SENSOR_REBOOT, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method prompts the base unit to send an *RST to the sensor in
/// HIFN  question. Measurements in progress are interrupted.
/// HIFN
/// HIFN     Attribute(s):
/// HIFN     RSPWRMETER_ATTR_CHANNEL_SENSOR_RESET
/// HIFN
/// HIFN     Remote-control command(s):
/// HIFN     SYSTem:SENSor<[1]..4>:RESet
/// HIFN
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_SensorReset(ViSession instrSession,
                                         ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_SENSOR_RESET,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_SENSOR_RESET, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Performs a sensor test and returns a list of strings separated by
/// HIFN  commas. The contents of this test protocol is sensor-specific. For its
/// HIFN  meaning, please refer to the sensor documentation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR arraySize/This control defines the size of array 'Test Result'.
/// HIPAR testResult/This control returns requested information from the sensor.
ViStatus _VI_FUNC rspwrmeter_SensorTest(ViSession instrSession,
                                        ViInt32   channel,
                                        ViInt32   arraySize,
                                        ViChar    testResult[])
{
    ViStatus  error = VI_SUCCESS;
    ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViChar*   response = NULL;
    ViBoolean isBaseUnit;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
    if (isBaseUnit)
    {
        viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
                2, "Channel");
        if (channel)
            snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SYST:SENS%ld:TEST?", channel);
        else
            strcpy(cmd, "SYST:SENS:TEST?");
    }
    else
    {
        strcpy(cmd, "TEST:SENS?");
    }

    checkErr(RsCore_QueryViStringUnknownLengthWithOpc(instrSession, cmd, 0, &response));
    RsCore_CopyToUserBufferAsciiData(testResult, arraySize, response);

Error:
    if (response) free(response);
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns information on a sensor.  For details refer to the
/// HIFN  documentation of the corresponding sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR arraySize/This control defines the size of array 'Info'.
/// HIPAR info/This control returns requested information from the sensor.
ViStatus _VI_FUNC rspwrmeter_SensorInfo(ViSession instrSession,
                                        ViInt32   channel,
                                        ViInt32   arraySize,
                                        ViChar    info[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_SENSOR_INFO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_SENSOR_INFO,
            arraySize, info),
            4, "Info");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  The filter length is automatically detected during the next
/// HIFN  measurement and set to that value. Thereafter automatic filter length
/// HIFN  detection is switched off. The setting can then be queried via Count
/// HIFN  attribute.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_AutoOnce(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_Write(instrSession, ":AVER:COUN:AUTO ONCE"));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Initializes the digital filter by deleting the stored measured values.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_AveragingReset(ViSession instrSession,
                                            ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_RESET,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_RESET, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  As soon as a new single value is determined, the filter window is
/// HIFN  advanced by one value so that the new value is taken into account by
/// HIFN  the filter and the oldest value is forgotten.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR terminalControl/Selects the terminal control.
ViStatus _VI_FUNC rspwrmeter_ConfigureAveragingTerminalControl(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   terminalControl)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_TERMINAL_CONTROL,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_TERMINAL_CONTROL, terminalControl),
            3, "Terminal Control");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the averaging method.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR averageDomain/Sets the averaging method.
ViStatus _VI_FUNC rspwrmeter_ConfigureAveragingDomain(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViInt32   averageDomain)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_DOMAIN,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_DOMAIN, averageDomain),
            3, "Average Domain");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Activates or deactivates the averaging and sets the filter length.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR averagingState/Activates or deactivates the filter function of a sensor.
/// HIPAR count/Sets the filter length i.e. the number of readings to be averaged for
/// HIPAR count/one measured value. The higher the count the lower the noise and the
/// HIPAR count/longer it takes to obtain a measured value.
ViStatus _VI_FUNC rspwrmeter_ConfigureAveragingCount(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean averagingState,
                                                     ViInt32   count)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_ENABLED, averagingState),
            3, "Averaging State");
    if (averagingState)
    {
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_COUNT, count),
                4, "Count");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the auto filter.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR mode/Selects the auto filter mode.
/// HIPAR refTimeslot/Sets a timeslot whose measured value is used to automatically
/// HIPAR refTimeslot/determine the filter length.
/// HIPAR noiseContent/Defines the proportion of intrinsic noise in the measured result (in
/// HIPAR noiseContent/dB). Specifically, this value gives the permitted relative deviation
/// HIPAR noiseContent/from the mean which may not be exceeded for 95% of the observation
/// HIPAR noiseContent/time.
/// HIPAR maxSettlingTime/Specifies an upper limit for the settling time in the "Fixed Noise"
/// HIPAR maxSettlingTime/mode.
/// HIPAR resolution/Defines the number of significant places for linear units and the
/// HIPAR resolution/number of decimal places for logarithmic units which should be free of
/// HIPAR resolution/noise in the measurement result.
ViStatus _VI_FUNC rspwrmeter_ConfigureAutoFilter(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   mode,
                                                 ViInt32   refTimeslot,
                                                 ViReal64  noiseContent,
                                                 ViReal64  maxSettlingTime,
                                                 ViInt32   resolution)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE, mode),
            3, "Mode");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_TIMESLOT, refTimeslot),
            4, "Ref Timeslot");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_NOISE_CONTENT, noiseContent),
            5, "Noise Content");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MTIME, maxSettlingTime),
            6, "Max Settling Time");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_RESOLUTION, resolution),
            7, "Resolution");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Sets auto filter to fixed noise mode and configures its parameters.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR noiseContent/Defines the proportion of intrinsic noise in the measured result (in
/// HIPAR noiseContent/dB). Specifically, this value gives the permitted relative deviation
/// HIPAR noiseContent/from the mean which may not be exceeded for 95% of the observation
/// HIPAR noiseContent/time.
/// HIPAR maxSettlingTime/Specifies an upper limit for the settling time in the "Fixed Noise"
/// HIPAR maxSettlingTime/mode.
ViStatus _VI_FUNC rspwrmeter_ConfigureAutoFilterNSR(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViReal64  noiseContent,
                                                    ViReal64  maxSettlingTime)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE, RSPWRMETER_VAL_AUTO_FILTER_MODE_FIXED_NOISE),
            1, "Mode");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_NOISE_CONTENT, noiseContent),
            3, "Noise Content");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MTIME, maxSettlingTime),
            4, "Max Settling Time");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Configures the auto filter to normal mode and sets the resolution.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR resolution/Defines the number of significant places for linear units and the
/// HIPAR resolution/number of decimal places for logarithmic units which should be free of
/// HIPAR resolution/noise in the measurement result.
ViStatus _VI_FUNC rspwrmeter_ConfigureAutoFilterResolution(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViInt32   resolution)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE, RSPWRMETER_VAL_AUTO_FILTER_MODE_NORMAL),
            1, "Mode");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_RESOLUTION, resolution),
            3, "Resolution");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Sets a timeslot whose measured value is used to automatically
/// HIFN determine the filter length. Available only for timeslot
/// HIFN measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR refTimeslot/Sets a timeslot whose measured value is used to automatically
/// HIPAR refTimeslot/determine the filter length.
ViStatus _VI_FUNC rspwrmeter_ConfigureAutoFilterReferenceTimeslot(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   refTimeslot)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_AVER_AUTO_TIMESLOT, refTimeslot),
            3, "Ref Timeslot");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Switches the offset correction on or off and sets the correction
/// HIFN  value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR correctionState/Switches the offset correction on or off.
/// HIPAR correction/A fixed offset value can be defined for multiplying (logarithmically
/// HIPAR correction/adding) the measured value of a sensor.
ViStatus _VI_FUNC rspwrmeter_ConfigureOffsetCorrection(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean correctionState,
                                                       ViReal64  correction)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_OFFSET_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CORR_OFFSET_ENABLED, correctionState),
            3, "Correction State");
    if (correctionState)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_OFFSET_VALUE, correction),
                4, "Correction");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Activates/deactivates the frequency-dependent offset correction
/// HIFN  specified in the selected table.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR tableState/Activates/deactivates the frequency-dependent offset correction
/// HIPAR tableState/specified in the selected table.
/// HIPAR selectedTable/Opens an item list containing correction tables.
ViStatus _VI_FUNC rspwrmeter_ConfigureTableOffsetCorrection(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViBoolean tableState,
                                                            ViString  selectedTable)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_OFFSET_TABLE_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CORR_OFFSET_TABLE_ENABLED, tableState),
            3, "Table State");

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CORR_OFFSET_SELECTED_TABLE, selectedTable),
            3, "Selected Table");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the frequency dependent offset. The frequency-dependent
/// HIFN  measured-value correction is set by selecting a table with the Selected
/// HIFN  Table property.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR FDOffset/Returns the maximum programmable current limit for the specified
/// HIPAR FDOffset/voltage level.
ViStatus _VI_FUNC rspwrmeter_QueryFDOffset(ViSession instrSession,
                                           ViInt32   channel,
                                           ViReal64* FDOffset)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_OFFSET_FD,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_OFFSET_FD, FDOffset),
            3, "FD Offset");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Instructs the sensor to perform a measured-value correction by means
/// HIFN  of one of the stored S-parameter tables (S-parameter device).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR sParameterState/Switches the offset correction on or off.
/// HIPAR correction/A fixed offset value can be defined for multiplying (logarithmically
/// HIPAR correction/adding) the measured value of a sensor.
ViStatus _VI_FUNC rspwrmeter_ConfigureSParameterDevice(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean sParameterState,
                                                       ViInt32   correction)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
    ViBoolean isBaseUnit;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_SPARAM_DEVICE_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
    if (isBaseUnit)
    {
        viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CORR_SPARAM_DEVICE_ENABLED, sParameterState),
                3, "S Parameter State");
    }
    if (sParameterState)
    {
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CORR_SPARAM_DEVICE, correction),
                4, "Correction");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the list of S parameter tables.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferSize/This control defines the size of buffer passed to Device List
/// HIPAR bufferSize/argument.
/// HIPAR deviceList/Returns the list of S parameter tables.
ViStatus _VI_FUNC rspwrmeter_GetSParameterDeviceList(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViInt32   bufferSize,
                                                     ViChar    deviceList[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_SPARAM_LIST,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CORR_SPARAM_LIST,
            bufferSize, deviceList),
            4, "Device List");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Activates/deactivates and configures the source gamma correction.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR correctionState/Activates/deactivates the measured-value correction of the reflection
/// HIPAR correctionState/coefficient effect of the source gamma.
/// HIPAR magnitude/This property sets the magnitude of the reflection coefficient.
/// HIPAR phase/This property sets the phase angle of the reflection coefficient.
ViStatus _VI_FUNC rspwrmeter_ConfigureSourceGammaCorrection(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViBoolean correctionState,
                                                            ViReal64  magnitude,
                                                            ViReal64  phase)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_SGAMMA_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CORR_SGAMMA_ENABLED, correctionState),
            3, "Correction State");
    if (correctionState)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_SGAMMA_MAGNITUDE, magnitude),
                4, "Magnitude");

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_SGAMMA_PHASE, phase),
                5, "Phase");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures reflection gamma correction.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR magnitude/This function sets the magnitude of the reflection coefficient of the
/// HIPAR magnitude/load for distortion compensation. To deactivate distortion
/// HIPAR magnitude/compensation, set Magnitude to 0. Unlike with mismatch correction,
/// HIPAR magnitude/there is no special activation/deactivation command. Reflection
/// HIPAR magnitude/compensation should remain deactivated if the values measured for the
/// HIPAR magnitude/load reflection coefficient are disputable.
/// HIPAR phase/Defines the phase angle (in degrees) of the complex reflection factor
/// HIPAR phase/of the load at the signal output.
ViStatus _VI_FUNC rspwrmeter_ConfigureReflectionGammaCorrection(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  magnitude,
                            ViReal64  phase)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CORR_RGAMMA_MAGNITUDE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_RGAMMA_MAGNITUDE, magnitude),
            3, "Magnitude");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CORR_RGAMMA_PHASE, phase),
            4, "Phase");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function enables or disables the duty cycle correction and sets
/// HIFN  the duty cycle correction for pulse power measurements.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR correctionState/Switches duty cycle correction for the measured-value on or off.
/// HIPAR correction/Informs the R&S NRP2 about the duty cycle of the power to be measured.
/// HIPAR correction/The duty cycle correction is only applied in the ContAv mode where
/// HIPAR correction/measurements are performed continuously without taking the timing of
/// HIPAR correction/the signal into account.
ViStatus _VI_FUNC rspwrmeter_ConfigureDutyCycleCorrection(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViBoolean correctionState,
                                                          ViReal64  correction)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_DCYCLE_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_DCYCLE_ENABLED, correctionState),
            3, "Correction State");
    if (correctionState)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_DCYCLE_VALUE, correction),
                4, "Correction");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Enables or disables the noise cancellation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR noiseCorrectionEnabled/Enables or disables the noise cancellation.
ViStatus _VI_FUNC rspwrmeter_ConfigureNoiseCorrection(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViBoolean noiseCorrectionEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_NOISE_CORRECTION_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_NOISE_CORRECTION_ENABLED, noiseCorrectionEnabled),
            3, "Noise Correction Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the magnitude, phase and expanded uncertainty of the complex
/// HIFN  input reflection coefficient Gamma_in.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omitt the channel numeric
/// HIPAR channel/suffix.
/// HIPAR magnitude/Queries the magnitude of the complex input reflection coefficient
/// HIPAR magnitude/Gamma_in.
/// HIPAR phase/Queries the phase angle of the complex input reflection coefficient
/// HIPAR phase/Gamma_in. The result is provided in degrees. Range: -180 degrees to
/// HIPAR phase/+180 degrees.
/// HIPAR expandedUncertainty/Queries the expanded (k = 2) uncertainty of the magnitude of the
/// HIPAR expandedUncertainty/complex input reflection coefficient Gamma_in. Following gamma
/// HIPAR expandedUncertainty/correction, the residual mismatch uncertainty becomes so small that it
/// HIPAR expandedUncertainty/is practically negligible.
ViStatus _VI_FUNC rspwrmeter_QueryIGammaCorrection(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64* magnitude,
                                                   ViReal64* phase,
                                                   ViReal64* expandedUncertainty)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_IGAMMA_MAGNITUDE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_IGAMMA_MAGNITUDE, magnitude),
            3, "Magnitude");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_IGAMMA_PHASE, phase),
            4, "Phase");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_IGAMMA_EXPANDED_UNCERTAINTY, expandedUncertainty),
            5, "Expanded Uncertainty");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  The "Cont Av" mode is the preferred measurement method if the
/// HIFN  measurement is not to be, or cannot be, synchronized with a specific
/// HIFN  signal event.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR aperture/Determines the integration time for a single measurement in the ContAv
/// HIPAR aperture/mode. To increase the measurement accuracy, this integration is
/// HIPAR aperture/followed by a second averaging procedure in a window with a selectable
/// HIPAR aperture/number of values.
/// HIPAR samplingRate/Sets the sampling rate.
/// HIPAR smoothingState/Activates/deactivates digital low pass filtering of the sampled video
/// HIPAR smoothingState/signal.
ViStatus _VI_FUNC rspwrmeter_ConfigureContAv(ViSession instrSession,
                                             ViInt32   channel,
                                             ViReal64  aperture,
                                             ViBoolean smoothingState,
                                             ViBoolean fastUnchoppedState)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
    ViBoolean isBaseUnit;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CONTAV_APERTURE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_APERTURE, aperture),
            3, "Aperture");

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_SMOOTHING_ENABLED, smoothingState),
            5, "Smoothing State");

    checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
    if (isBaseUnit)
    {
        viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_FAST_UNCHOPPED_ENABLED, fastUnchoppedState),
                6, "Fast Unchopped State");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the Cont Av buffer state and size.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferState/Switches on the buffered ContAv mode, after which data blocks rather
/// HIPAR bufferState/than single measured values are then returned. In this mode a higher
/// HIPAR bufferState/data rate is achieved than in the non-buffered ContAv mode.
/// HIPAR bufferSize/Sets the number of desired values for the buffered ContAv mode.
ViStatus _VI_FUNC rspwrmeter_ConfigureContAvBuffer(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViBoolean bufferState,
                                                   ViInt32   bufferSize)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_ENABLED, bufferState),
            3, "Buffer State");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_SIZE, bufferSize),
            4, "Buffer Size");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the buffer count.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferCount/Returns the buffer count.
ViStatus _VI_FUNC rspwrmeter_GetContAvBufferCount(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32*  bufferCount)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_COUNT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_COUNT, bufferCount),
            3, "Buffer Count");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Clears the contents of the continuous average result buffer.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_ClearContAvBuffer(ViSession instrSession,
                                               ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_CLEAR,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CONTAV_BUFFER_CLEAR, NULL),
            3, "Buffer Clear");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the BurstAv dropout tolerance, Exclude Start and Stop.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR dropoutTolerance/Defines the end of the burst. If power keeps low for at least this
/// HIPAR dropoutTolerance/time, the end of the burst is assumed. Modulation-specific power drops
/// HIPAR dropoutTolerance/that are shorter than the set value are ignored.
/// HIPAR excludeStart/Sets a time that is to be excluded at the beginning of the integration
/// HIPAR excludeStart/period.
/// HIPAR excludeStop/Sets a time that is to be excluded at the end of the integration
/// HIPAR excludeStop/period.
ViStatus _VI_FUNC rspwrmeter_ConfigureBurstAv(ViSession instrSession,
                                              ViInt32   channel,
                                              ViReal64  dropoutTolerance,
                                              ViReal64  excludeStart,
                                              ViReal64  excludeStop)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_BURSTAV_DROPOUT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_BURSTAV_DROPOUT, dropoutTolerance),
            3, "Dropout Tolerance");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_BURSTAV_EXCLUDE_START, excludeStart),
            4, "Exclude Start");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_BURSTAV_EXCLUDE_STOP, excludeStop),
            5, "Exclude Stop");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the length of the burst.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR length/Returns the length of the burst.
ViStatus _VI_FUNC rspwrmeter_GetBurstAvLength(ViSession instrSession,
                                              ViInt32   channel,
                                              ViReal64* length)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_BURSTAV_LENGTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_BURSTAV_LENGTH, length),
            3, "Length");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the duration of the sampling window. During this unsynchronized
/// HIFN  time interval, the average signal power is measured.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR aclrApertureTime/Sets the duration of the sampling window. During this unsynchronized
/// HIPAR aclrApertureTime/time interval, the average signal power is measured.
ViStatus _VI_FUNC rspwrmeter_ConfigureACLRApertureTime(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64  aclrApertureTime)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_ACLR_APERTURE_TIME,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_ACLR_APERTURE_TIME, aclrApertureTime),
            3, "ACLR Aperture Time");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the distance from transmission channel to adjacent channel in
/// HIFN  the ACLR mode. The return value depends on the transmission standard
/// HIFN  selected by the resolution bandwidth setting.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR aclrSpacing/Queries the distance from transmission channel to adjacent channel in
/// HIPAR aclrSpacing/the ACLR mode. The return value depends on the transmission standard
/// HIPAR aclrSpacing/selected by the resolution bandwidth setting.
ViStatus _VI_FUNC rspwrmeter_GetACLRSpacing(ViSession instrSession,
                                            ViInt32   channel,
                                            ViReal64* aclrSpacing)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_ACLR_SPACING,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_ACLR_SPACING, aclrSpacing),
            3, "ACLR Spacing");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Set/Query IQ trace result sample rate (requires OPT-K1)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR voltIQSampleRate/Set/Query IQ trace result sample rate (requires OPT-K1)
ViStatus _VI_FUNC rspwrmeter_ConfigureVoltIQSampleRate(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64  voltIQSampleRate)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_VOLT_IQ_SAMPLE_RATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_VOLT_IQ_SAMPLE_RATE, voltIQSampleRate),
            3, "Volt IQ Sample Rate");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  IQ trace result sample count (requires OPT-K1)
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR voltIQSampleCount/IQ trace result sample count (requires OPT-K1)
ViStatus _VI_FUNC rspwrmeter_ConfigureVoltIQSampleCount(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViInt32   voltIQSampleCount)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_VOLT_IQ_SAMPLE_COUNT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_VOLT_IQ_SAMPLE_COUNT, voltIQSampleCount),
            3, "Volt IQ Sample Count");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR time/Sets the time to be covered by the Trace sequence.
/// HIPAR traceOffsetTime/Determines the beginning of the trace measurement sequence relative to
/// HIPAR traceOffsetTime/the trigger event.
/// HIPAR realtime/This control is reserved.
/// HIPAR points/Sets the number of desired values per Trace sequence.
ViStatus _VI_FUNC rspwrmeter_ConfigureTrace(ViSession instrSession,
                                            ViInt32   channel,
                                            ViReal64  time,
                                            ViReal64  traceOffsetTime,
                                            ViBoolean reserved,
                                            ViInt32   points)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";
             reserved;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_TIME,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_TIME, time),
            3, "Time");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_OFFSET_TIME, traceOffsetTime),
            4, "Trace Offset Time");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_POINTS, points),
            6, "Points");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the trace time automatically.
/// HIFN     Performs an FFT to calculate the power spectrum and sets the trace
/// HIFN  time to five times the larges peak of the spectrum, which is a good
/// HIFN  indication for the dominant period of the signal.
/// HIFN     This calculation is performed for any signal without checking the
/// HIFN  signal-to-noise ratio.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR traceTimeAuto/Sets the trace time automatically.
ViStatus _VI_FUNC rspwrmeter_ConfigureTraceTimeAuto(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean traceTimeAuto)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_TRACE_TIME_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_TRACE_TIME_AUTO, traceTimeAuto),
            3, "Trace Time Auto");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the lower and upper limit for the power axis in the Trace mode.
/// HIFN  The value range depends on the unit.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR lowerLimit/Sets the lower limit for the power axis in the Trace mode. The value
/// HIPAR lowerLimit/range depends on the unit.
/// HIPAR upperLimit/Sets the upper limit for the power axis in the Trace mode.
ViStatus _VI_FUNC rspwrmeter_ConfigureTraceLimit(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViReal64  lowerLimit,
                                                 ViReal64  upperLimit)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_LOWER_LIMIT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_UPPER_LIMIT, upperLimit),
            4, "Upper Limit");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_LOWER_LIMIT, lowerLimit),
            3, "Lower Limit");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries minimum slot width.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR minimumSlotWidth/Queries minimum slot width.
ViStatus _VI_FUNC rspwrmeter_GetTraceMinimumSlotWidth(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64* minimumSlotWidth)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MINIMUM_SLOT_WIDTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MINIMUM_SLOT_WIDTH, minimumSlotWidth),
            3, "Minimum Slot Width");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Number of samples in trace result
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR numberOfSamples/Number of samples in trace result
ViStatus _VI_FUNC rspwrmeter_GetNumberOfSamples(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32*  numberOfSamples)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_NUMBER_OF_SAMPLES,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_NUMBER_OF_SAMPLES, numberOfSamples),
            3, "Number Of Samples");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sample rate for trace
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR sampleRateForTrace/Sample rate for trace
ViStatus _VI_FUNC rspwrmeter_GetSampleRateForTrace(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViReal64* sampleRateForTrace)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_SAMPLE_RATE_FOR_TRACE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_SAMPLE_RATE_FOR_TRACE, sampleRateForTrace),
            3, "Sample Rate For Trace");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  The automatic switchover is deactivated and the suitable filter length
/// HIFN  is then defined and set automatically.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_ResetAverageCount(ViSession instrSession,
                                               ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS%ld:TRAC:AVER:COUN:AUTO ONCE", channel);
    else
        strcpy(cmd, "SENS:TRAC:AVER:COUN:AUTO ONCE");
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the filter function for the trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR averagingState/Switches the filter function of a sensor on or off for the Trace mode.
/// HIPAR countAuto/Can be used to automatically determine a value for property Average
/// HIPAR countAuto/Count.
/// HIPAR count/Sets the length of the filter for the Trace mode. The wider the
/// HIPAR count/filter, the lower the noise and the longer it takes to obtain a
/// HIPAR count/measured value.
/// HIPAR terminalControl/Sets the terminal control.
ViStatus _VI_FUNC rspwrmeter_ConfigureTraceAveraging(ViSession instrSession,
                                                     ViInt32   channel,
                                                     ViBoolean averagingState,
                                                     ViBoolean countAuto,
                                                     ViInt32   count,
                                                     ViInt32   terminalControl)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    repCap[RS_REPCAP_BUF_SIZE] = "";
    ViBoolean isBaseUnit;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_FILTER_AVER_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_FILTER_AVER_ENABLED, averagingState),
            3, "Averaging State");
    if (averagingState)
    {
        checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
        if (isBaseUnit)
        {
            viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_FILTER_AVER_COUNT_AUTO, countAuto),
                    4, "Count Auto");
        }
        if (!countAuto)
        {
            viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_FILTER_AVER_COUNT, count),
                    5, "Count");
        }
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_FILTER_TERMINAL_CONTROL, terminalControl),
                6, "Terminal Control");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the trace measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR measurementState/Switches on/off automatic pulse measurements in the Trace mode. If
/// HIPAR measurementState/pulse measurement is on, the sensor automatically determines the pulse
/// HIPAR measurementState/parameters for the currently measured trace.
/// HIPAR algorithm/Selects the analysis algorithm.
/// HIPAR time/This property sets the time.
/// HIPAR timeOffset/This property sets the offset time.
/// HIPAR autoTransfer/Switches on/off automatic transfer of the measured pulse parameters
/// HIPAR autoTransfer/after each trace. This enables a synchronous display of the trace and
/// HIPAR autoTransfer/the pulse data during a continuous measurement.
ViStatus _VI_FUNC rspwrmeter_ConfigureTraceMeasurement(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViBoolean measurementState,
                                                       ViInt32   algorithm,
                                                       ViReal64  time,
                                                       ViReal64  timeOffset,
                                                       ViBoolean autoTransfer)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_ALGORITHM,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_ENABLED, measurementState),
            3, "Measurement State");
    if (measurementState)
    {
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_ALGORITHM, algorithm),
                4, "Algorithm");

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TIME, time),
                5, "Time");

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_OFFSET_TIME, timeOffset),
                6, "Time Offset");

        viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_AUTO_TRANSFER, autoTransfer),
                7, "Auto Transfer");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the trace transition measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR referenceLevel/Sets the medial reference level in terms of percentage of the pulse
/// HIPAR referenceLevel/power amplitude. This level is used to define the pulse width, pulse
/// HIPAR referenceLevel/start time and pulse stop time.
/// HIPAR highReference/Sets the high reference level in terms of percentage of the pulse
/// HIPAR highReference/power amplitude. The "High Ref. Power" defines the end of the rising
/// HIPAR highReference/edge and the start of the falling edge of the pulse. These values are
/// HIPAR highReference/needed for measurement of the rise / fall time.
/// HIPAR lowReference/Sets the low reference level in terms of percentage of the pulse power
/// HIPAR lowReference/amplitude. The "Low Ref. Power" defines the start of the rising edge
/// HIPAR lowReference/and the end of the falling edge of the pulse. These values are needed
/// HIPAR lowReference/for measurement of the rise / fall time.
ViStatus _VI_FUNC rspwrmeter_ConfigureTraceTransitionMeasurement(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  referenceLevel,
                            ViReal64  highReference,
                            ViReal64  lowReference)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_REFERENCE_LEVEL,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_REFERENCE_LEVEL, referenceLevel),
            3, "Reference Level");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_HIGH_REFERENCE, highReference),
            4, "High Reference");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_LOW_REFERENCE, lowReference),
            3, "Low Reference");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the trace transition measurement results.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR edge/Selects the rising or falling edge.
/// HIPAR duration/Returns the rise or fall time of the first detected pulse.
/// HIPAR occurrence/Returns the start or stop point of the current pulse, i.e. the time
/// HIPAR occurrence/when the signal passes through the medial reference level with rising
/// HIPAR occurrence/or falling edge.
/// HIPAR overshoot/Returns the amount of relative undershooting of the signal at the
/// HIPAR overshoot/beginning or end of a pulse.
ViStatus _VI_FUNC rspwrmeter_QueryTraceTransitionMeasurement(ViSession instrSession,
                                                             ViInt32   channel,
                                                             ViInt32   edge,
                                                             ViReal64* duration,
                                                             ViReal64* occurrence,
                                                             ViReal64* overshoot)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_POS_DURATION,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    switch (edge)
    {
    case RSPWRMETER_VAL_TRACE_MEAS_EDGE_RISE:
        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_POS_DURATION, duration),
                4, "Duration");

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_POS_OCCURRENCE, occurrence),
                5, "Occurrence");

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_POS_OVERSHOOT, overshoot),
                6, "Overshoot");
        break;
    case RSPWRMETER_VAL_TRACE_MEAS_EDGE_FALL:
        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_NEG_DURATION, duration),
                4, "Duration");

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_NEG_OCCURRENCE, occurrence),
                5, "Occurrence");

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_TRANS_NEG_OVERSHOOT, overshoot),
                6, "Overshoot");
        break;
    default:
        viCheckParm(RsCore_InvalidViInt32Value(instrSession, edge), 3, "Edge");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the trace pulse measurement results.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR dutyCycle/Returns the duty cycle of the current trace data.
/// HIPAR width/Returns the pulse duration (pulse width) of the first detected pulse
/// HIPAR width/in the current trace.
/// HIPAR period/Returns the time the pulse signal needs to complete one cycle.
/// HIPAR separation/Returns the time from "Stop Time" to "Start Time" of the first
/// HIPAR separation/detected pulse.
ViStatus _VI_FUNC rspwrmeter_QueryTracePulseMeasurement(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViReal64* dutyCycle,
                                                        ViReal64* width,
                                                        ViReal64* period,
                                                        ViReal64* separation)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_PULSE_DUTY_CYCLE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_PULSE_DUTY_CYCLE, dutyCycle),
            3, "Duty Cycle");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_PULSE_WIDTH, width),
            4, "Width");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_PULSE_PERIOD, period),
            5, "Period");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_PULSE_SEPARATION, separation),
            6, "Separation");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the trace power measurement results.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR average/Returns the measured average power in the current trace.
/// HIPAR maximum/Returns the measured maximum power in the current trace.
/// HIPAR minimum/Returns the measured minimum power in the current trace.
/// HIPAR basePower/Returns the base power, computed in the pulse level analysis histogram
/// HIPAR basePower/of the current trace data.
/// HIPAR topPower/Returns the power of the first detected complete pulse of the current
/// HIPAR topPower/trace data, including rising and falling edges.
ViStatus _VI_FUNC rspwrmeter_QueryTracePowerMeasurement(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViReal64* average,
                                                        ViReal64* maximum,
                                                        ViReal64* minimum,
                                                        ViReal64* basePower,
                                                        ViReal64* topPower)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_AVERAGE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_AVERAGE, average),
            3, "Average");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_MAXIMUM, maximum),
            4, "Maximum");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_MINIMUM, minimum),
            5, "Minimum");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_BASE, basePower),
            6, "Base Power");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_TOP, topPower),
            7, "Top Power");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Returns the droop of the pulse. The droop is the difference
///HINF  between leading edge amplitude, pulse top, and falling-edge
///HINF  amplitude. The droop is expressed in percentage of the pulse top
///HINF  amplitude.
///HIRET Returns the status code of this operation.
///HIPAR channel/Numerical selection of a channel. Use 0 to omitt the
///HIPAR channel/channel numeric suffix.
///HIPAR droop/Returns the droop of the pulse. The droop is the
///HIPAR droop/difference between leading edge amplitude, pulse top, and
///HIPAR droop/falling-edge amplitude. The droop is expressed in percentage
///HIPAR droop/of the pulse top amplitude.
ViStatus _VI_FUNC rspwrmeter_QueryTracePulseDroop(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViReal64* droop)
{
    ViStatus error = VI_SUCCESS;
    ViChar   channelRepCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_DROOP,
        "Channel", channel, RS_REPCAP_BUF_SIZE, channelRepCap), 2, "Channel");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, channelRepCap, RSPWRMETER_ATTR_CHANNEL_TRACE_MEAS_POWER_DROOP, droop),
            3, "Droop");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the timegate (depends on trigger event) in
/// HIFN  which the sensor is doing statistic measurements.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR offset/This property sets the start after the trigger of the timegate in
/// HIPAR offset/which the sensor is doing statistic measurements.
/// HIPAR time/This property sets the length of the timegate in which the sensor is
/// HIPAR time/doing statistic measurements.
ViStatus _VI_FUNC rspwrmeter_ConfigureStatisticsTimegate(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViReal64  offset,
                                                         ViReal64  time)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_TIME,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_OFFSET_TIME, offset),
            3, "Offset Time");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_TIME, time),
            4, "Time");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the timegate fence (depends on trigger event)
/// HIFN  in which the sensor is doing statistic measurements.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR midambleOffset/This property sets the midamble offset after timeslot start in seconds
/// HIPAR midambleOffset/in the timegate in which the sensor is doing statistic measurements.
/// HIPAR midambleLength/This property sets the midamble length in seconds.
ViStatus _VI_FUNC rspwrmeter_ConfigureStatisticsTimegateFence(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64  midambleOffset,
                            ViReal64  midambleLength)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_MID_LENGTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_MID_LENGTH, midambleLength),
            4, "Midamble Length");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_MID_OFFSET, midambleOffset),
            3, "Midamble Offset");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function sets the X-Axis of statistical measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR referenceLevel/Defines, together with the property Scale Range, the range on the power
/// HIPAR referenceLevel/axis over which the statistics function (CCDF/PDF) is calculated.
/// HIPAR range/Defines the range on the power axis over which the statistics function
/// HIPAR range/(CCDF/PDF) is calculated.
/// HIPAR points/Determines the number of samples on the power axis for the Statistics
/// HIPAR points/modes or returns the current setting. Due to the limited storage
/// HIPAR points/capacity of the R&S NRP2, the number is limited to 1024.
ViStatus _VI_FUNC rspwrmeter_ConfigureStatisticsScale(ViSession instrSession,
                                                      ViInt32   channel,
                                                      ViReal64  referenceLevel,
                                                      ViReal64  range,
                                                      ViInt32   points)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_RLEVEL,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_RLEVEL, referenceLevel),
            3, "Reference Level");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_RANGE, range),
            4, "Range");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_POINTS, points),
            5, "Points");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Enters the required minimum number of samples.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR samples/Enters the required minimum number of samples.
ViStatus _VI_FUNC rspwrmeter_ConfigureStatisticsMinimumSamples(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32   samples)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_SAMPLES,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_SAMPLES, samples),
            3, "Samples");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures the statistics measurements (CCDF and PDF).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR power/This property sets the horizontally shiftable marker on the power
/// HIPAR power/axis. The absolute position is specified.
/// HIPAR verticalPosition/This property sets the vertically shiftable marker on the statistics
/// HIPAR verticalPosition/axis.
ViStatus _VI_FUNC rspwrmeter_ConfigureStatisticsMeasurement(ViSession instrSession,
                                                            ViInt32   channel,
                                                            ViReal64  power,
                                                            ViReal64  verticalPosition)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_POWER,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_POWER, power),
            3, "Power");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_VERTICAL_POSITION, verticalPosition),
            4, "Vertical Position");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the values of the statistics function.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR horizontalData/This property returns the value of the statistics function at the
/// HIPAR horizontalData/position of the marker if the value is valid.
/// HIPAR absolutePowerPosition/Returns the absolute position of the marker on the power axis.
/// HIPAR peakPower/Returns the peak power value calculated during a statistics
/// HIPAR peakPower/measurement (CCDF/PDF).
/// HIPAR average/Returns the average power value calculated during a statistics
/// HIPAR average/measurement (CCDF/PDF).
ViStatus _VI_FUNC rspwrmeter_QueryStatisticsMeasurement(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViReal64* horizontalData,
                                                        ViReal64* absolutePowerPosition,
                                                        ViReal64* peakPower,
                                                        ViReal64* average)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_HORIZONTAL_DATA,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_HORIZONTAL_DATA, horizontalData),
            3, "Horizontal Data");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_ABS_POWER_POSITION, absolutePowerPosition),
            4, "Absolute Power Position");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_PEAK_POWER, peakPower),
            5, "Peak Power");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_AVERAGE, average),
            6, "Average");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the minimum width of a sample on the power axis.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR minimumWidth/Returns the minimum width of a sample on the power axis.
ViStatus _VI_FUNC rspwrmeter_QueryStatisticsScaleMinimumWidth(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64* minimumWidth)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_MIN_WIDTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_STAT_SCALE_MIN_WIDTH, minimumWidth),
            3, "Minimum Width");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the average value of the power in the time window of a
/// HIFN  statistics measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR averageData/Returns the average value of the power in the time window of a
/// HIPAR averageData/statistics measurement.
ViStatus _VI_FUNC rspwrmeter_ReadStatisticsAverageData(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64* averageData)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC%ld:STAT:POW:AVG:DATA?", channel);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC:STAT:POW:AVG:DATA?");

    checkErr(rspwrmeter_QueryViReal64(instrSession, cmd, averageData));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the parameters of Timeslot measurement mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR timeSlotCount/Sets the number of simultaneously measured timeslots.
/// HIPAR width/Defines the timeslot length.
ViStatus _VI_FUNC rspwrmeter_ConfigureTimeslot(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   timeSlotCount,
                                               ViReal64  width)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_COUNT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_COUNT, timeSlotCount),
            3, "Time Slot Count");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_WIDTH, width),
            4, "Width");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function sets the time interval at the beginning and end of the
/// HIFN  burst that is excluded from the measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR excludeStart/Sets a time that is to be excluded at the beginning of the integration
/// HIPAR excludeStart/period.
/// HIPAR excludeStop/Sets a time that is to be excluded at the end of the integration
/// HIPAR excludeStop/period.
ViStatus _VI_FUNC rspwrmeter_ConfigureTimeslotInterval(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViReal64  excludeStart,
                                                       ViReal64  excludeStop)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_EXCLUDE_START,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_EXCLUDE_START, excludeStart),
            3, "Exclude Start");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_EXCLUDE_STOP, excludeStop),
            4, "Exclude Stop");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Activates and configures the "Fence" function.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR midambleState/Activates or deactivates the blanking out of time intervals in the
/// HIPAR midambleState/timeslots.
/// HIPAR midambleOffset/This property sets sets the start of an exclusion interval in a
/// HIPAR midambleOffset/timeslot.
/// HIPAR midambleLength/This control sets the length of an exclusion interval in a timeslot.
ViStatus _VI_FUNC rspwrmeter_ConfigureTimeslotFence(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean midambleState,
                                                    ViReal64  midambleOffset,
                                                    ViReal64  midambleLength)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_MID_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_MID_ENABLED, midambleState),
            3, "Midamble State");
    if (midambleState)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_MID_OFFSET, midambleOffset),
                4, "Midamble Offset");

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_TIMESLOT_MID_LENGTH, midambleLength),
                5, "Midamble Length");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects one of the four gates in which the average power is measured
/// HIFN  if the Timegate mode is selected.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR timegate/Selects one of the four gates.
ViStatus _VI_FUNC rspwrmeter_SelectedTimegate(ViSession instrSession,
                                              ViInt32   channel,
                                              ViInt32   timegate)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_SELECTED_TIMEGATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_SELECTED_TIMEGATE, timegate),
            3, "Selected Timegate");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the timegate.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR timegate/Selects one of the four gates.
/// HIPAR offset/This property sets the start of the timegate after the trigger.
/// HIPAR time/This property sets the length of a timegate for recording measured
/// HIPAR time/values in the Timegate Average mode.
/// HIPAR midambleState/Activates/deactivates the blanking out of time intervals in the
/// HIPAR midambleState/timegates, affects all gates.
/// HIPAR midambleOffset/This property is used to define the start of an exclusion interval in
/// HIPAR midambleOffset/the specified timegate. The start of the exclusion interval is
/// HIPAR midambleOffset/referenced to the start of the timegate.
/// HIPAR midambleLength/This property is used to specify the length of an exclusion interval.
ViStatus _VI_FUNC rspwrmeter_ConfigureTimegate(ViSession instrSession,
                                               ViInt32   channel,
                                               ViInt32   timegate,
                                               ViReal64  offset,
                                               ViReal64  time,
                                               ViBoolean midambleState,
                                               ViReal64  midambleOffset,
                                               ViReal64  midambleLength)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";
    ViChar   repCap_timegate[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_GATE_TIME,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_GATE_TIME,
            "Timegate", timegate- 1, RS_REPCAP_BUF_SIZE, repCap_timegate));

    strcat(repCap, ",");
    strcat(repCap, repCap_timegate);

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_GATE_TIME, time),
            5, "Time");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_GATE_OFFSET_TIME, offset),
            4, "Offset Time");

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_GATE_MID_ENABLED, midambleState),
            6, "Midamble State");

    if (midambleState)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_GATE_MID_OFFSET, midambleOffset),
                7, "Midamble Offset");

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_GATE_MID_LENGTH, midambleLength),
                8, "Midamble Length");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This attribute switches the optional RF generator (ON) or (OFF).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR refOscillatorState/This attribute switches the optional RF generator (ON) or (OFF).
ViStatus _VI_FUNC rspwrmeter_ConfigureRefOscillatorEnabled(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean refOscillatorState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_REF_OSC_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_REF_OSC_ENABLED, refOscillatorState),
            3, "Ref Oscillator State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures reference oscillator source and frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR referenceOscillatorSource/Reference oscillator source.
/// HIPAR referenceOscillatorFrequency/Sets the frequency of the reference clock signal that is supplied at
/// HIPAR referenceOscillatorFrequency/the REF connector, CLK connector or USB connector according to selected
/// HIPAR referenceOscillatorFrequency/source.
ViStatus _VI_FUNC rspwrmeter_ConfigureReferenceOscillator(ViSession instrSession,
                                                          ViInt32   referenceOscillatorSource,
                                                          ViReal64  referenceOscillatorFrequency)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, NULL, RSPWRMETER_ATTR_REFERENCE_OSCILLATOR_SOURCE, referenceOscillatorSource),
            2, "Reference Oscillator Source");
    switch (referenceOscillatorSource)
    {
    case RSPWRMETER_VAL_ROSC_SOURCE_USB:
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, NULL, RSPWRMETER_ATTR_REFERENCE_OSCILLATOR_USB_FREQUENCY, referenceOscillatorFrequency),
                3, "Reference Oscillator USB Frequency");
        break;
    case RSPWRMETER_VAL_ROSC_SOURCE_REF:
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, NULL, RSPWRMETER_ATTR_REFERENCE_OSCILLATOR_REF_FREQUENCY, referenceOscillatorFrequency),
                3, "Reference Oscillator REF Frequency");
        break;
    case RSPWRMETER_VAL_ROSC_SOURCE_CLK:
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, NULL, RSPWRMETER_ATTR_REFERENCE_OSCILLATOR_CLK_FREQUENCY, referenceOscillatorFrequency),
                3, "Reference Oscillator CLK Frequency");
        break;
    case RSPWRMETER_VAL_ROSC_SOURCE_INTERNAL:
    default:
        break;
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  If the REF connector is used as an output, enables or disables the
/// HIFN  output signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR referenceIOOutputEnabled/If the REF connector is used as an output, enables or disables the
/// HIPAR referenceIOOutputEnabled/output signal.
/// HIPAR sampleClockOutputEnabled/Enable sample clock output on external connector.
ViStatus _VI_FUNC rspwrmeter_ConfigureSMAConnectorOutput(ViSession instrSession,
                                                         ViBoolean referenceIOOutputEnabled,
                                                         ViBoolean sampleClockOutputEnabled)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, NULL, RSPWRMETER_ATTR_REFERENCE_IO_OUTPUT_ENABLED, referenceIOOutputEnabled),
            2, "Reference IO Output Enabled");

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, NULL, RSPWRMETER_ATTR_SAMPLE_CLOCK_OUTPUT_ENABLED, sampleClockOutputEnabled),
            3, "Sample Clock Output Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Write binary calibration data sets for the selected sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR dataBlock/Selects the type of data block.
/// HIPAR arraySize/This control defines the size of array 'Data'.
/// HIPAR data/This control writes binary calibration data.
ViStatus _VI_FUNC rspwrmeter_WriteCalibrationData(ViSession instrSession,
                                                  ViInt32   channel,
                                                  ViInt32   dataBlock,
                                                  ViInt32   arraySize,
                                                  ViChar    data[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViString data_blockArr[] = {"", "USER:"};

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, dataBlock, RSPWRMETER_VAL_DATA_BLOCK_FACTORY, RSPWRMETER_VAL_DATA_BLOCK_USER),
            3, "Data Block");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CAL%ld:%sDATA,", channel, data_blockArr[dataBlock]);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CAL:%sDATA,", data_blockArr[dataBlock]);
    checkErr(RsCore_WriteBinaryDataBlock(instrSession, cmd, (ViByte*)data, arraySize));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Read binary calibration data sets for the selected sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/This control defines the channel number.
/// HIPAR dataBlock/Selects the type of data block.
/// HIPAR arraySize/This control defines the size of array 'Data'.
/// HIPAR data/This control writes binary calibration data.
ViStatus _VI_FUNC rspwrmeter_ReadCalibrationData(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViInt32   dataBlock,
                                                 ViInt32   arraySize,
                                                 ViChar    data[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViByte*  p2buf = NULL;
    ViInt64  count = 0;
    ViString data_blockArr[] = {"", "USER:"};

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, dataBlock, RSPWRMETER_VAL_DATA_BLOCK_FACTORY, RSPWRMETER_VAL_DATA_BLOCK_USER),
            3, "Data Block");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CAL%ld:%sDATA?", channel, data_blockArr[dataBlock]);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CAL:%sDATA?", data_blockArr[dataBlock]);
        checkErr(RsCore_QueryBinaryDataBlock(instrSession, cmd, VI_TRUE, NULL, &p2buf, &count));

    checkErr(rspwrmeter_CheckStatus(instrSession));

    checkErr(RsCore_CopyToUserBufferBinData((ViBuf)data, arraySize, p2buf, (ViInt32)count));

Error:
    if (p2buf) free(p2buf);
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method first does an internal heater test with CAL:TEST and
/// HIFN  returns the relative deviation between the test result and the result
/// HIFN  that was measured in the calibration lab during sensor calibration.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR testDeviation/This control returns the relative deviation between the test result
/// HIPAR testDeviation/and the result that was measured in the calibration lab during sensor
/// HIPAR testDeviation/calibration
ViStatus _VI_FUNC rspwrmeter_GetTestDeviation(ViSession instrSession,
                                              ViInt32   channel,
                                              ViReal64* testDeviation)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CAL_DEVIATION,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CAL_DEVIATION, testDeviation),
            3, "Test Deviation");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method returns the heater test result that was measured in the
/// HIFN  calibration lab during sensor calibration.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR testReference/This control returns the heater test result that was measured in the
/// HIPAR testReference/calibration lab during sensor calibration.
ViStatus _VI_FUNC rspwrmeter_GetTestReference(ViSession instrSession,
                                              ViInt32   channel,
                                              ViReal64* testReference)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CAL_REFERENCE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_CAL_REFERENCE, testReference),
            3, "Test Reference");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the length of calibration data block.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR dataBlock/Selects the type of data block.
/// HIPAR length/Length of calibration data block.
ViStatus _VI_FUNC rspwrmeter_GetDataBlockLength(ViSession instrSession,
                                                ViInt32   channel,
                                                ViInt32   dataBlock,
                                                ViInt32*  length)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";
    ViAttr   attr[2] = {
             RSPWRMETER_ATTR_CHANNEL_CAL_DATA_BLOCK_LENGTH,
             RSPWRMETER_ATTR_CHANNEL_CAL_USER_DATA_BLOCK_LENGTH
             };

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, dataBlock, RSPWRMETER_VAL_DATA_BLOCK_FACTORY, RSPWRMETER_VAL_DATA_BLOCK_USER),
            3, "Data Block");

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_CAL_DATA_BLOCK_LENGTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, attr[dataBlock], length),
            4, "Length");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Performs the automatic settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
ViStatus _VI_FUNC rspwrmeter_Autoset(ViSession instrSession,
                                     ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_AUTOSET_FILTER_TYPE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    checkErr(rspwrmeter_AutosetFrequency(instrSession, channel));
    checkErr(rspwrmeter_AutosetInputAttenuation(instrSession, channel));
    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_AUTOSET_FILTER_TYPE, VI_TRUE),
            1, "Autoset Filter Type");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Adjusts center frequency and bandwidth settings to the test signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
ViStatus _VI_FUNC rspwrmeter_AutosetFrequency(ViSession instrSession,
                                              ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_AUTOSET_FREQUENCY,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_AUTOSET_FREQUENCY, NULL),
            3, "Autoset Frequency");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}


/// HIFN  Adjusts the input attenuation one time, then disables the automatic
/// HIFN  setting.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
ViStatus _VI_FUNC rspwrmeter_AutosetInputAttenuation(ViSession instrSession,
                                                     ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_AUTOSET_INPUT_ATTENUATION,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_AUTOSET_INPUT_ATTENUATION, NULL),
            3, "Autoset Input Attenuation");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Select bandwidth from resolution bandwidth or from required sample
/// HIFN  rate.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR bandwidthType/Select bandwidth from resolution bandwidth or from required sample
/// HIPAR bandwidthType/rate.
ViStatus _VI_FUNC rspwrmeter_ConfigureBandwidthType(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViInt32   bandwidthType)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_BANDWIDTH_TYPE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_BANDWIDTH_TYPE, bandwidthType),
            3, "Bandwidth Type");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Resolution bandwidth for power measurement. Only discrete steps are
/// HIFN  available. The entered value is rounded to the next value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR resolutionBandwidth/Resolution bandwidth for power measurement. Only discrete steps are
/// HIPAR resolutionBandwidth/available. The entered value is rounded to the next value.
ViStatus _VI_FUNC rspwrmeter_ConfigureResolutionBandwidth(ViSession instrSession,
                                                          ViInt32   channel,
                                                          ViReal64  resolutionBandwidth)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_RESOLUTION_BANDWIDTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_RESOLUTION_BANDWIDTH, resolutionBandwidth),
            3, "Resolution Bandwidth");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Currently used value of resolution bandwidth.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR currentResolutionBandwidth/Currently used value of resolution bandwidth.
ViStatus _VI_FUNC rspwrmeter_GetCurrentResolutionBandwidth(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViReal64* currentResolutionBandwidth)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CURRENT_RESOLUTION_BANDWIDTH,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_CURRENT_RESOLUTION_BANDWIDTH, currentResolutionBandwidth),
            3, "Current Resolution Bandwidth");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the filter type for resolution bandwidth filter. The filter
/// HIFN  bandwidth is not affected.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR autosetFilterType/If enabled, sets the filter type suitable for the currently chosen
/// HIPAR autosetFilterType/measurement mode and bandwidth.
/// HIPAR filterType/Sets the filter type for resolution bandwidth filter. The filter
/// HIPAR filterType/bandwidth is not affected.
ViStatus _VI_FUNC rspwrmeter_ConfigureFilterType(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean autosetFilterType,
                                                 ViInt32   filterType)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_AUTOSET_FILTER_TYPE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_AUTOSET_FILTER_TYPE, autosetFilterType),
            3, "Autoset Filter Type");
    if (autosetFilterType == VI_FALSE)
    {
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_FILTER_TYPE, filterType),
                4, "Filter Type");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures input attenuation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR inputAttenuationAuto/Automatic mode for input attenuation.
/// HIPAR inputAttenuation/Fixed input attenuation. Only two values are possible, 0 dB and 30 dB.
/// HIPAR inputAttenuation/The entered value is rounded to the next value.
ViStatus _VI_FUNC rspwrmeter_ConfigureInputAttenuationAuto(ViSession instrSession,
                                                           ViInt32   channel,
                                                           ViBoolean inputAttenuationAuto,
                                                           ViReal64  inputAttenuation)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_INPUT_ATTENUATION_AUTO,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_INPUT_ATTENUATION_AUTO, inputAttenuationAuto),
            3, "Input Attenuation Auto");

    if (inputAttenuationAuto == VI_FALSE)
    {
        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_INPUT_ATTENUATION, inputAttenuation),
                4, "Input Attenuation");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Enables or disables the output of the local oscillator signal.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR oscillatorOutputEnabled/Enables or disables the output of the local oscillator signal.
ViStatus _VI_FUNC rspwrmeter_ConfigureLocalOscillatorOutput(ViSession instrSession,
                                                            ViBoolean oscillatorOutputEnabled)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, NULL, RSPWRMETER_ATTR_LOCAL_OSCILLATOR_OUTPUT_ENABLED, oscillatorOutputEnabled),
            2, "Oscillator Output Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Query the current local oscillator frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR localOscillatorFrequency/Query the current local oscillator frequency.
ViStatus _VI_FUNC rspwrmeter_GetLocalOscillatorFrequency(ViSession instrSession,
                                                         ViInt32   channel,
                                                         ViReal64* localOscillatorFrequency)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_IF_SIDEBAND_AUTO_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_LOCAL_OSCILLATOR_FREQUENCY, localOscillatorFrequency),
            3, "Local Oscillator Frequency");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures intermediate frequency sideband.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR ifSidebandAutoEnabled/Enables or disables the automatic setting of the intermediate
/// HIPAR ifSidebandAutoEnabled/frequency sideband.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR ifSidebandSelected/Sets the currently used intermediate frequency sideband.
ViStatus _VI_FUNC rspwrmeter_ConfigureIFSideband(ViSession instrSession,
                                                 ViInt32   channel,
                                                 ViBoolean ifSidebandAutoEnabled,
                                                 ViInt32   ifSidebandSelected)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_IF_SIDEBAND_AUTO_ENABLED,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_IF_SIDEBAND_AUTO_ENABLED, ifSidebandAutoEnabled),
            3, "IF Sideband Auto Enabled");
    if (ifSidebandAutoEnabled == VI_FALSE)
    {
        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_IF_SIDEBAND_SELECTED, ifSidebandSelected),
                4, "IF Sideband Selected");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the currently used intermediate frequency.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR ifSidebandFrequency/Queries the currently used intermediate frequency.
ViStatus _VI_FUNC rspwrmeter_GetIFSidebandFrequency(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViReal64* ifSidebandFrequency)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_IF_SIDEBAND_FREQUENCY,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_IF_SIDEBAND_FREQUENCY, ifSidebandFrequency),
            3, "IF Sideband Frequency");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Enables or disables the frequency tracker.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR frequencyTrackingEnabled/Enables or disables the frequency tracker.
ViStatus _VI_FUNC rspwrmeter_ConfigureFrequencyTracking(ViSession instrSession,
                                                        ViInt32   channel,
                                                        ViBoolean frequencyTrackingEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_FREQUENCY_TRACKING_ENABLED, frequencyTrackingEnabled),
            3, "Frequency Tracking Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the current state of the frequency tracker.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR trackingCurrentState/Queries the current state of the frequency tracker.
ViStatus _VI_FUNC rspwrmeter_GetFrequencyTrackingCurrentState(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViInt32*  trackingCurrentState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_FREQUENCY_TRACKING_CURRENT_STATE, trackingCurrentState),
            3, "Tracking Current State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the current frequency tracking position. This frequency is
/// HIFN  estimated based on the 10 MHz reference signal and is assigned to the
/// HIFN  test signal. It is used to determine the power measurement value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omit the channel numeric
/// HIPAR channel/suffix.
/// HIPAR trackingCurrentFrequency/Queries the current frequency tracking position. This frequency is
/// HIPAR trackingCurrentFrequency/estimated based on the 10 MHz reference signal and is assigned to the
/// HIPAR trackingCurrentFrequency/test signal. It is used to determine the power measurement value.
ViStatus _VI_FUNC rspwrmeter_GetFrequencyTrackingCurrentFrequency(
                            ViSession instrSession,
                            ViInt32   channel,
                            ViReal64* trackingCurrentFrequency)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MODE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_FREQUENCY_TRACKING_CURRENT_FREQUENCY, trackingCurrentFrequency),
            3, "Tracking Current Frequency");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Performs zeroing using the signal at the sensor input(s). The
/// HIFN  sensor(s) must be disconnected from all power sources.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_Zero(ViSession instrSession,
                                  ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_ZERO_ONCE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ZERO_ONCE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Perform fast zeroing. However, the commands cannot be called unless
/// HIFN  the sensor(s) is(are) in the Trace mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_FastZero(ViSession instrSession,
                                      ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_FAST_ZERO_ONCE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_FAST_ZERO_ONCE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method queries return the value ON if a calibration is in
/// HIFN  progress, otherwise value OFF.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR zeroingComplete/This control returns the value ON if a calibration is in progress,
/// HIPAR zeroingComplete/otherwise value OFF.
ViStatus _VI_FUNC rspwrmeter_IsZeroComplete(ViSession  instrSession,
                                            ViInt32    channel,
                                            ViBoolean* zeroingComplete)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_ZERO_STATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_ZERO_STATE, zeroingComplete),
            3, "Zeroing Complete");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method initiates a signal acquisition based on the present
/// HIFN  instrument configuration. It then waits for the acquisition to
/// HIFN  complete, and returns the measured value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR timeout/Pass the maximum length of time in which to allow the read measurement
/// HIPAR timeout/operation to complete.
/// HIPAR measurement/Returns the measured value.
ViStatus _VI_FUNC rspwrmeter_ReadMeasurement(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   timeout,
                                             ViReal64* measurement)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "READ%ld?", channel);
    else
        strcpy(cmd, "READ?");

    checkErr(RsCore_QueryViReal64WithOpc(instrSession, cmd, timeout, measurement));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method returns the last valid measurement result. The method is
/// HIFN  only completed after a valid measurement has been performed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR measurement/Returns the measured value.
/// HIPAR bufferSize/Defines the size of buffer passed to 'Measurement'.
/// HIPAR returnedValues/Returns count of results read from instrument and stored in buffer passed to 'Measurement'.
ViStatus _VI_FUNC rspwrmeter_FetchMeasurement(ViSession instrSession,
                                              ViInt32   channel,
                                              ViUInt32  bufferSize,
                                              ViReal64  measurement[],
                                              ViUInt32* returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld?", channel);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC?");

    checkErr(RsCore_QueryFloatArrayToUserBuffer(instrSession, cmd, bufferSize, measurement, (ViInt32 *)returnedValues));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method returns the last valid IQ samples (only usable for NRQ6).
/// HIFN  The method is only completed after a valid measurement has been performed.
/// HiFN  Use the rspwrmeter_MeasurementMode to set the IQ data Acquisition.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR iData/Returns the I data.
/// HIPAR qData/Returns the Q data.
/// HIPAR bufferSize/Defines the size of the buffers passed to 'Measurement'.
/// HIPAR returnedValues/Returns count of results read from instrument and stored in buffer passed to 'Measurement'.
ViStatus _VI_FUNC rspwrmeter_FetchIqData(ViSession instrSession,
                                         ViInt32   channel,
                                         ViUInt32  bufferSize,
                                         ViReal64  iData[],
                                         ViReal64  qData[],
                                         ViUInt32* returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViChar   oldFmt[RS_MAX_MESSAGE_BUF_SIZE];
    ViReal64 *data = NULL;
    ViInt32  dataSize, count, i, ix;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    checkErr(RsCore_QueryViString(instrSession, "FORM:DATA?", oldFmt));
    checkErr(RsCore_Write(instrSession, "FORM:DATA REAL,32"));

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld?", channel);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC?");

    checkErr(RsCore_QueryFloatArray(instrSession, cmd, &data, &dataSize));
    count = dataSize / 2;

    if (returnedValues)
        *returnedValues = (ViUInt32)count;

    if (count > (ViInt32)bufferSize)
        count = (ViInt32)bufferSize;

    ix = 0;
    for (i = 0; i < count; i++)
    {
        iData[i] = data[ix++];
        qData[i] = data[ix++];
    }

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FORM:DATA %s", oldFmt);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    if (data)
        free(data);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the measurement results of all channels of a sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferSize/Defines the size of buffer passed to 'Measurement'.
/// HIPAR measurement/Returns the measured value.
/// HIPAR returnedValues/Returns count of results read from instrument and stored in buffer passed to 'Measurement'.
ViStatus _VI_FUNC rspwrmeter_FetchAllPaths(ViSession instrSession,
                                           ViInt32   channel,
                                           ViUInt32  bufferSize,
                                           ViReal64  measurement[],
                                           ViUInt32* returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld:ALL?", channel);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:ALL?");

    checkErr(RsCore_QueryFloatArrayToUserBufferWithOpc(instrSession, cmd, 0, bufferSize, measurement, (ViInt32 *)returnedValues));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the last valid measurement result of a measurement with enabled data buffer mode.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR bufferSize/Defines the size of buffer passed to 'Measurement'.
/// HIPAR measurement/Returns the measured value.
/// HIPAR returnedValues/Returns count of results read from instrument and stored in buffer passed to 'Measurement'.
ViStatus _VI_FUNC rspwrmeter_FetchAllPathsBuffered(ViSession instrSession,
                                                   ViInt32   channel,
                                                   ViUInt32  bufferSize,
                                                   ViReal64  measurement[],
                                                   ViUInt32* returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
    {
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld:ALL:FORM BLOC", channel);
        checkErr(RsCore_Write(instrSession, cmd));
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld:ARR?", channel);
        checkErr(RsCore_QueryFloatArrayToUserBufferWithOpc(instrSession, cmd, 0, bufferSize, measurement, (ViInt32 *)returnedValues));
    }
    else
    {
        checkErr(RsCore_Write(instrSession, "FETC:ALL:FORM BLOC")); // IGNORE: Check double write
        checkErr(RsCore_QueryFloatArrayToUserBufferWithOpc(instrSession, "FETC:ARR?", 0, bufferSize, measurement, (ViInt32 *)returnedValues));
    }

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Triggers a BUS event. If the sensor is in the WAIT_FOR_TRG state and
/// HIFN  the source for the trigger source is set to BUS, the sensor enters the
/// HIFN  MEASURING state. This function invalidates all current measuring
/// HIFN  results. A query of measurement data following this function will thus
/// HIFN  always return the measured value determined in response to this
/// HIFN  function.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_SendSoftwareTrigger(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_MEAS_SEND_SOFTWARE_TRIGGER, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the current measured value as the reference value.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
ViStatus _VI_FUNC rspwrmeter_ResetRelativeMeasurement(ViSession instrSession,
                                                      ViInt32   calculateBlock)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, calculateBlock, 1, 8),
            2, "Calculate Block");
    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "CALC%ld:REL:AUTO ONCE", calculateBlock);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Configures Relative measurement settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR state/Relates a power or a power ratio to a reference value.
/// HIPAR reference/Determines a value that is used as a divisor (logarithmic subtraction)
/// HIPAR reference/for all measured values of the enabled calculate block.
/// HIPAR referencePowerRatio/Has the same direct effect as property Reference on the reference
/// HIPAR referencePowerRatio/value for measured values that share the unit of a power ratio.
ViStatus _VI_FUNC rspwrmeter_ConfigureRelativeMeasurement(ViSession instrSession,
                                                          ViInt32   calculateBlock,
                                                          ViBoolean state,
                                                          ViReal64  reference,
                                                          ViReal64  referencePowerRatio)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_RELATIVE_ENABLED,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_RELATIVE_ENABLED, state),
            3, "State");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_RELATIVE_REFERENCE, reference),
            4, "Reference");

    viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_RELATIVE_REFERENCE_POWER_RATIO, referencePowerRatio),
            5, "Reference Power Ratio");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects a measurement function that processes one sensor or two
/// HIFN  together.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR expression/This control may assume the values given in the following table, n and
/// HIPAR expression/m may assume 1, 2, 3 and 4. A sensor must be connected to the
/// HIPAR expression/respective channel.
ViStatus _VI_FUNC rspwrmeter_ConfigureMathExpression(ViSession instrSession,
                                                     ViInt32   calculateBlock,
                                                     ViString  expression)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_MATH_EXPRESSION,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_MEAS_MATH_EXPRESSION, expression),
            3, "Expression");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Lists all supported calculation functions. All functions are sent in
/// HIFN  the form of strings which are allowed as parameters for the Expression
/// HIFN  attribute.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR bufferSize/Pass the number of bytes in the string you specify for the catalog
/// HIPAR bufferSize/list parameter.
/// HIPAR mathExpressionList/This control lists all supported calculation functions.
ViStatus _VI_FUNC rspwrmeter_GetMathExpressionCatalog(ViSession instrSession,
                                                      ViInt32   calculateBlock,
                                                      ViInt32   bufferSize,
                                                      ViChar    mathExpressionList[])
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_MATH_EXPRESSION_CATALOG,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_MATH_EXPRESSION_CATALOG,
            bufferSize, mathExpressionList),
            4, "mathExpressionList");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the position of a curve marker on the time axis.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR marker/Numerical selection of a marker
/// HIPAR xPosition/This control sets the position of a curve marker on the time axis.
ViStatus _VI_FUNC rspwrmeter_SetTraceMarkerXPosition(ViSession instrSession,
                                                     ViInt32   calculateBlock,
                                                     ViInt32   marker,
                                                     ViReal64  xPosition)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, marker, 1, 2),
            3, "marker");

    if (marker == 1)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION1,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION1, xPosition),
                4, "xPosition");
    }
    else if (marker == 2)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION2,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_SetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION2, xPosition),
                4, "xPosition");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the position of a curve marker on the time axis.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR marker/Numerical selection of a marker
/// HIPAR xPosition/This control returns the position of a curve marker on the time axis.
ViStatus _VI_FUNC rspwrmeter_GetTraceMarkerXPosition(ViSession instrSession,
                                                     ViInt32   calculateBlock,
                                                     ViInt32   marker,
                                                     ViReal64* xPosition)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, marker, 1, 2),
            3, "marker");

    if (marker == 1)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION1,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION1, xPosition),
                4, "xPosition");
    }
    else if (marker == 2)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION2,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_XPOSITION2, xPosition),
                4, "xPosition");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the position of a curve marker on the power axis if the value
/// HIFN  is valid.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR marker/Numerical selection of a marker
/// HIPAR yPosition/This control returns the position of a curve marker on the power axis
/// HIPAR yPosition/if the value is valid.
ViStatus _VI_FUNC rspwrmeter_GetTraceMarkerYPosition(ViSession instrSession,
                                                     ViInt32   calculateBlock,
                                                     ViInt32   marker,
                                                     ViReal64* yPosition)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, marker, 1, 2),
            3, "marker");

    if (marker == 1)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_YPOSITION1,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_YPOSITION1, yPosition),
                4, "yPosition");
    }
    else if (marker == 2)
    {
        checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_YPOSITION2,
                "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

        viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_YPOSITION2, yPosition),
                4, "yPosition");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  It returns the time and power difference of the two curve markers if
/// HIFN  the value is valid.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR xDelta/It returns the time difference of the two curve markers.
/// HIPAR yDelta/Returns the power difference of the two curve markers.
ViStatus _VI_FUNC rspwrmeter_QueryTraceMarkerDelta(ViSession instrSession,
                                                   ViInt32   calculateBlock,
                                                   ViReal64* xDelta,
                                                   ViReal64* yDelta)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_TRACE_XDELTA,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_XDELTA, xDelta),
            3, "xDelta");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_TRACE_YDELTA, yDelta),
            4, "yDelta");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the displayed values from the results window.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR calculateBlock/Numerical selection of a calculate block
/// HIPAR dataCount/Queries the number of measured values that are included when the mean
/// HIPAR dataCount/value and standard deviation are calculated (in the above formulas
/// HIPAR dataCount/referred to as N, see Mean value of all measured values and Standard
/// HIPAR dataCount/deviation of all measured values).
/// HIPAR data/With this command the current output value of a calculate block can be
/// HIPAR data/queried.
/// HIPAR standardDeviation/Queries the standard deviation of all measured values of the calculate
/// HIPAR standardDeviation/block. The standard deviation is calculated together with the mean
/// HIPAR standardDeviation/value and reset.
/// HIPAR mean/Queries the mean value of all measured values of the calculate block.
/// HIPAR mean/The mean value is reset when the auxiliary values are reset.
/// HIPAR peakToPeak/Queries the peak-to-peak distance (maximum to minimum) of the measured
/// HIPAR peakToPeak/values of the calculate block.
/// HIPAR maximum/Queries the maximum of all measured values that previously occurred in
/// HIPAR maximum/this calculate block.
/// HIPAR minimum/Queries the minimum of all measured values that previously occurred in
/// HIPAR minimum/this calculate block.
ViStatus _VI_FUNC rspwrmeter_QueryDisplayedData(ViSession instrSession,
                                                ViInt32   calculateBlock,
                                                ViInt32*  dataCount,
                                                ViReal64* data,
                                                ViReal64* standardDeviation,
                                                ViReal64* mean,
                                                ViReal64* peakToPeak,
                                                ViReal64* maximum,
                                                ViReal64* minimum)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEAS_DISPLAY_DATA_COUNT,
            "CalculateBlock", calculateBlock- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_DATA_COUNT, dataCount),
            3, "dataCount");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_DATA, data),
            4, "data");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_SDEVIATION, standardDeviation),
            5, "standardDeviation");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_MEAN, mean),
            6, "mean");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_PEAK_TO_PEAK, peakToPeak),
            8, "peakToPeak");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_MAXIMUM, maximum),
            8, "maximum");

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, repCap, RSPWRMETER_ATTR_MEAS_DISPLAY_MINIMUM, minimum),
            9, "minimum");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Immediately sets the respective sensors to the IDLE state.
/// HIFN  Measurements in progress are interrupted.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_AbortAll(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_Write(instrSession, "ABOR:ALL"));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method starts a measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_InitiateAll(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_Write(instrSession, "INIT:ALL"));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method starts a measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR timeout/Pass the maximum length of time in which to allow the read measurement operation to complete.
ViStatus _VI_FUNC rspwrmeter_InitiateAllAndWait(ViSession instrSession,
                                                ViInt32   timeout)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_WriteWithOpc(instrSession, "INIT:ALL", timeout));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function prevents the execution of the Initiate property for all
/// HIFN  sensors.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR disableInitiate/This control prevents the execution of the Initiate property for all
/// HIPAR disableInitiate/sensors.
ViStatus _VI_FUNC rspwrmeter_DisableInitiateAll(ViSession instrSession,
                                                ViBoolean disableInitiate)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, "", RSPWRMETER_ATTR_MEAS_DISABLE_INITIATE, disableInitiate),
            2, "disableInitiate");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects either single-shot or continuous (free-running) measurement
/// HIFN  cycles.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR continuousState/Selects either single-shot or continuous (free-running) measurement
/// HIPAR continuousState/cycles.
ViStatus _VI_FUNC rspwrmeter_SetInitContinuousStateAll(ViSession instrSession,
                                                       ViBoolean continuousState)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, "", RSPWRMETER_ATTR_MEAS_CONTINUOUS, continuousState),
            2, "Continuous State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Queries for channels that are connected to an antenna.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR sensorNumber/Selects the sensor.
/// HIPAR sensorPresent/Queries for channels that are connected to an antenna.
ViStatus _VI_FUNC rspwrmeter_QueryMeasPathSensorPresent(ViSession  instrSession,
                                                        ViInt32    sensorNumber,
                                                        ViBoolean* sensorPresent)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEASUREMENT_PATH_SENSOR_PRESENT,
            "Path", sensorNumber- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_MEASUREMENT_PATH_SENSOR_PRESENT, sensorPresent),
            3, "Sensor Present");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Activates channels that are connected to an antenna.
/// HIRET Returns the status code of this operation.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR sensorNumber/Selects the sensor.
/// HIPAR sensorEnabled/Activates channels that are connected to an antenna.
ViStatus _VI_FUNC rspwrmeter_ConfigureMeasPathSensorEnabled(ViSession instrSession,
                                                            ViInt32   sensorNumber,
                                                            ViBoolean sensorEnabled)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_MEASUREMENT_PATH_SENSOR_ENABLED,
            "Path", sensorNumber- 1, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_MEASUREMENT_PATH_SENSOR_ENABLED, sensorEnabled),
            3, "Sensor Enabled");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Immediately sets the respective sensors to the IDLE state.
/// HIFN  Measurements in progress are interrupted.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_Abort(ViSession instrSession,
                                   ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_ABORT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_ABORT, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method starts a measurement.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
ViStatus _VI_FUNC rspwrmeter_Initiate(ViSession instrSession,
                                      ViInt32   channel)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_INITIATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_INITIATE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This method starts a measurement and waits for it to finish.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR timeout/Pass the maximum length of time in which to allow the read measurement operation to complete.
ViStatus _VI_FUNC rspwrmeter_InitiateAndWait(ViSession instrSession,
                                             ViInt32   channel,
                                             ViInt32   timeout)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";
    ViInt32  old_timeout = 0;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_INITIATE_WAIT,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));
    if (timeout > 0)
    {
        checkErr(rspwrmeter_GetOPCTimeout(instrSession, &old_timeout));
        checkErr(rspwrmeter_SetOPCTimeout(instrSession, timeout));
    }

    checkErr(rspwrmeter_SetAttributeViString(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_INITIATE_WAIT, NULL));

Error:
    if (old_timeout > 0)
    {
        (void)rspwrmeter_SetOPCTimeout(instrSession, old_timeout);
    }

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This property prevents the execution of the Initiate property for one
/// HIFN  or more sensors.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR disableInitiate/This property prevents the execution of the Initiate property for one
/// HIPAR disableInitiate/or more sensors.
ViStatus _VI_FUNC rspwrmeter_DisableInitiate(ViSession instrSession,
                                             ViInt32   channel,
                                             ViBoolean disableInitiate)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_DISABLE_INITIATE,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_DISABLE_INITIATE, disableInitiate),
            3, "Disable Initiate");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Selects either single-shot or continuous (free-running) measurement
/// HIFN  cycles.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR continuousState/Selects either single-shot or continuous (free-running) measurement
/// HIPAR continuousState/cycles.
ViStatus _VI_FUNC rspwrmeter_SetInitContinuousState(ViSession instrSession,
                                                    ViInt32   channel,
                                                    ViBoolean continuousState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_CONTINUOUS,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_SetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_CONTINUOUS, continuousState),
            3, "Continuous State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function returns whether single-shot or continuous (free-running)
/// HIFN  measurement is selected.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR continuousState/This control returns the state of continuous initiate.
ViStatus _VI_FUNC rspwrmeter_GetInitContinuousState(ViSession  instrSession,
                                                    ViInt32    channel,
                                                    ViBoolean* continuousState)
{
    ViStatus error = VI_SUCCESS;
    ViChar   repCap[RS_REPCAP_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, RSPWRMETER_ATTR_CHANNEL_MEAS_CONTINUOUS,
            "Channel", channel, RS_REPCAP_BUF_SIZE, repCap));

    viCheckParm(rspwrmeter_GetAttributeViBoolean(instrSession, repCap, RSPWRMETER_ATTR_CHANNEL_MEAS_CONTINUOUS, continuousState),
            3, "Continuous State");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the trace data.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR arraySize/Number of array points requested.
/// HIPAR traceData/A user allocated buffer into which the trace data is stored.
/// HIPAR returnedValues/Number of points actually returned in the array.
ViStatus _VI_FUNC rspwrmeter_GetTraceData(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   arraySize,
                                          ViReal64  traceData[],
                                          ViInt32*  returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViByte*  data = NULL;
    ViInt64  count = 0;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS%ld:TRAC:DATA?", channel);
    else
        strcpy(cmd, "SENS:TRAC:DATA?");

    checkErr(RsCore_QueryBinaryDataBlock(instrSession, cmd, VI_TRUE, NULL, &data, &count));
    checkErr(rspwrmeter_CheckStatus(instrSession));

    checkErr(rspwrmeter_ParseUserDataContent(instrSession, (ViChar *)data, (ViInt32)count, arraySize, traceData, NULL,
        returnedValues, NULL, NULL));

Error:
    if (data) free(data);
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Returns the trace data.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR arraySize/Number of array points requested.
/// HIPAR expectedDataSize/Document parameter here.
/// HIPAR traceData/A user allocated buffer into which the trace data is stored.
/// HIPAR resultType/Array of returned Result Types. For each section one value.
/// HIPAR returnedDataSize/Array of number of points actually returned in the Trace Data
/// HIPAR returnedDataSize/array. For each section one value.
/// HIPAR returnedSections/Document parameter here.
ViStatus _VI_FUNC rspwrmeter_GetTraceDataWithAuxiliary(ViSession instrSession,
                                                       ViInt32   channel,
                                                       ViInt32   arraySize,
                                                       ViInt32   expectedDataSize,
                                                       ViReal64  traceData[],
                                                       ViInt32   resultType[],
                                                       ViInt32   returnedDataSize[],
                                                       ViInt32*  returnedSections)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViByte*   tracData = NULL;
    ViInt64   count = 0;
    ViReal64* p2traceData = traceData;
    ViReal64* tmpArr = NULL;
    ViInt32   tmpResultType;
    ViInt32   tmpReturnedDataSize;
    ViInt32   tmpCount;
    ViChar*   tmpData;
    ViInt32   tmpCopyCount = 0;
    ViInt32   i = 0;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS%ld:TRAC:DATA?", channel);
    else
        strcpy(cmd, "SENS:TRAC:DATA?");

    checkErr(RsCore_QueryBinaryDataBlock(instrSession, cmd, VI_TRUE, VI_FALSE, &tracData, &count));
    checkErr(rspwrmeter_CheckStatus(instrSession));
    tmpData = (ViChar*)tracData;
    tmpCount = (ViInt32)count;
    tmpArr = malloc(sizeof(ViReal64) * expectedDataSize);
    while (tmpCount > 0)
    {
        rspwrmeter_ParseUserDataContent(instrSession, tmpData, tmpCount, expectedDataSize, tmpArr,
                                        &tmpResultType, &tmpReturnedDataSize, &tmpData, &tmpCount);
        if (i < arraySize)
        {
            //copy data
            tmpCopyCount = (tmpReturnedDataSize < expectedDataSize) ? tmpReturnedDataSize : expectedDataSize;
            memcpy(p2traceData, tmpArr, sizeof(ViReal64) * tmpCopyCount);
            p2traceData += tmpCopyCount;
            //copy result type
            if (resultType) resultType[i] = tmpResultType;
            //copy data size
            if (returnedDataSize) returnedDataSize[i] = tmpReturnedDataSize;
        }
        i++;
    }
    if (returnedSections) *returnedSections = i;

Error:
    if (tracData) free(tracData);
    if (tmpArr) free(tmpArr);
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Queries the results of the continuous average result buffer and returns them even if the buffer is not full.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel
/// HIPAR arraySize/Number of array points requested.
/// HIPAR bufferData/A user allocated buffer into which the data is stored.
/// HIPAR returnedValues/Number of points actually returned in the array.
ViStatus _VI_FUNC rspwrmeter_GetBufferData(ViSession instrSession,
                                           ViInt32   channel,
                                           ViInt32   arraySize,
                                           ViReal64  bufferData[],
                                           ViInt32*  returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS%ld:POW:AVG:BUFF:DATA?", channel);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SENS:POW:AVG:BUFF:DATA?");

    checkErr(RsCore_QueryFloatArrayToUserBuffer(instrSession, cmd, arraySize ,bufferData, returnedValues));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the last valid measurement result.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR channel/Numerical selection of a channel. Use 0 to omitt the channel numeric
/// HIPAR channel/suffix.
/// HIPAR measurement/Selects the measurement to be returned.
/// HIPAR arraySize/Number of array points requested.
/// HIPAR results/A user allocated buffer into which the measurement data is stored.
/// HIPAR returnedValues/Number of points actually returned in the array.
ViStatus _VI_FUNC rspwrmeter_FetchResults(ViSession instrSession,
                                          ViInt32   channel,
                                          ViInt32   measurement,
                                          ViInt32   arraySize,
                                          ViReal64  results[],
                                          ViInt32*  returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar*  p2rng;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, measurement, RSPWRMETER_VAL_MEAS_CONT_AVER, RSPWRMETER_VAL_MEAS_CONT_TIMESLOT),
            3, "Measurement");
    viCheckParm(RsCore_FindRangeTableDiscreteCmdString(instrSession, &RsPwrMeter_rngFetchResult, measurement, &p2rng),
                3, "Measurement");

    if (channel)
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC%ld:SCAL:POW:%s?", channel, p2rng);
    else
        snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "FETC:SCAL:POW:%s?", p2rng);
        checkErr(RsCore_QueryFloatArrayToUserBuffer(instrSession, cmd, arraySize, results, returnedValues));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN This function sets the Enable, NTransition, and PTransition bits of selected registers.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR registerGroup/Selects the register group.
/// HIPAR enable/Sets the enable bits of the status register.
/// HIPAR PTransition/Sets the PTransition bits (edge detectors) of the status register.
/// HIPAR NTransition/Sets the NTransition bits (edge detectors) of the status register.
ViStatus _VI_FUNC rspwrmeter_SetStatusRegister(ViSession instrSession,
                                               ViInt32   registerGroup,
                                               ViUInt16  enable,
                                               ViUInt16  PTransition,
                                               ViUInt16  NTransition)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, registerGroup, RSPWRMETER_VAL_REGISTER_GROUP_OPER, RSPWRMETER_VAL_REGISTER_GROUP_QUES_WIND),
            2, "Register Group");

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "STAT:%s:ENAB %hu;PTR %hu;NTR %hu", gsc_StatusRegArr[registerGroup], enable, PTransition,
            NTransition);
    checkErr(RsCore_Write(instrSession, cmd));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN This function contains commands for querying the status reporting system.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR registerGroup/Selects the register group.
/// HIPAR registerPart/Selects the register part.
/// HIPAR registerValue/Returns content of selected register.
ViStatus _VI_FUNC rspwrmeter_GetStatusRegister(ViSession instrSession,
                                               ViInt32   registerGroup,
                                               ViInt32   registerPart,
                                               ViUInt16* registerValue)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";
    ViInt32  value;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, registerGroup, RSPWRMETER_VAL_REGISTER_GROUP_OPER, RSPWRMETER_VAL_REGISTER_GROUP_QUES_WIND),
            2, "Register Group");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, registerGroup, RSPWRMETER_VAL_REGISTER_TYPE_COND, RSPWRMETER_VAL_REGISTER_TYPE_EVEN),
            3, "Register Part");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, registerValue),
        4, "Register Value");

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "STAT:%s:%s?", gsc_StatusRegArr[registerGroup], gsc_RegTypeArr[registerPart]);
    checkErr(rspwrmeter_QueryViInt32(instrSession, cmd, &value));
    *registerValue = (ViInt16)value;

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Clears all bits in the Questionable Data enable register and the Standard Operation enable register.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_PresetStatusRegisters(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_STATUS_PRESET, NULL),
            2, "Preset");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Function used by Hi-level functions to obtain a repCap name from:
/// HIFN - attributeID
/// HIFN - repCapNameID
/// HIFN - 0-based index in the repCapNames
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR attributeId/Pass the ID of an attribute.
/// HIPAR repCapNameId/RepCap Name ID.
/// HIPAR index/Zero based index in the repCapNames.
/// HIPAR bufferSize/Pass the number of bytes in the string you allocate for the repCapName.
/// HIPAR repCapName/String returning RepCap name
ViStatus _VI_FUNC rspwrmeter_GetAttributeRepCapName(ViSession instrSession,
                                                    ViUInt32  attributeId,
                                                    ViString  repCapNameId,
                                                    ViInt32   index,
                                                    ViInt32   bufferSize,
                                                    ViChar    repCapName[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetAttributeRepCapName(instrSession, attributeId, repCapNameId, index, bufferSize, repCapName));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_SetAttribute<type> Functions
* Purpose:  These functions enable the instrument driver user to set
*           attribute values directly.  There are type-safe versions for
*           ViInt32, ViReal64, ViString, ViBoolean, and ViSession data types.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_SetAttributeViInt32(ViSession instrSession,
                                                 ViString  repeatedCapabilityName,
                                                 ViUInt32  attributeId,
                                                 ViInt32   value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_SetAttributeViInt32(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_SetAttributeViInt64(ViSession instrSession,
                                                 ViString  repeatedCapabilityName,
                                                 ViAttr    attributeId,
                                                 ViInt64   value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_SetAttributeViInt64(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_SetAttributeViReal64(ViSession instrSession,
                                                  ViString  repeatedCapabilityName,
                                                  ViUInt32  attributeId,
                                                  ViReal64  value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_SetAttributeViReal64(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_SetAttributeViString(ViSession instrSession,
                                                  ViString  repeatedCapabilityName,
                                                  ViUInt32  attributeId,
                                                  ViString  value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_SetAttributeViString(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_SetAttributeViBoolean(ViSession instrSession,
                                                   ViString  repeatedCapabilityName,
                                                   ViUInt32  attributeId,
                                                   ViBoolean value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_SetAttributeViBoolean(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_GetAttribute<type> Functions
* Purpose:  These functions enable the instrument driver user to get
*           attribute values directly.  There are type-safe versions for
*           ViInt32, ViReal64, ViString, ViBoolean, and ViSession attributes.
*****************************************************************************/

ViStatus _VI_FUNC rspwrmeter_GetAttributeViInt32(ViSession instrSession,
                                                 ViString  repeatedCapabilityName,
                                                 ViUInt32  attributeId,
                                                 ViInt32*  value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);
    error = RsCore_GetAttributeViInt32(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_GetAttributeViInt64(ViSession instrSession,
                                                 ViString  repeatedCapabilityName,
                                                 ViAttr    attributeId,
                                                 ViInt64*  value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_GetAttributeViInt64(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_GetAttributeViReal64(ViSession instrSession,
                                                  ViString  repeatedCapabilityName,
                                                  ViUInt32  attributeId,
                                                  ViReal64* value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_GetAttributeViReal64(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_GetAttributeViString(ViSession instrSession,
                                                  ViString  repeatedCapabilityName,
                                                  ViUInt32  attributeId,
                                                  ViInt32   bufferSize,
                                                  ViChar    value[])
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_GetAttributeViString(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, bufferSize, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

ViStatus _VI_FUNC rspwrmeter_GetAttributeViBoolean(ViSession  instrSession,
                                                   ViString   repeatedCapabilityName,
                                                   ViUInt32   attributeId,
                                                   ViBoolean* value)
{
    ViStatus error = VI_SUCCESS;

    (void)RsCore_LockSession(instrSession);

    error = RsCore_GetAttributeViBoolean(instrSession, repeatedCapabilityName, attributeId, RS_VAL_DIRECT_USER_CALL, value);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}


/*****************************************************************************
* Function: Rs_GetAttributeRepCapNameId
* Purpose:  This function returns the repeated capability id(s) that
*           belongs to the attribute you specify. More then one id is listed
*           in comma separated string.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_GetAttributeRepeatedCapabilityIds(
                            ViSession instrSession,
                            ViUInt32  attributeId,
                            ViInt32   bufferSize,
                            ViChar    repeatedCapabilityIds[])
{
    ViStatus error = VI_SUCCESS;

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, bufferSize, 1, RS_MAX_SHORT_MESSAGE_BUF_SIZE),
            3, "Buffer Size");

    checkErr(RsCore_GetAttributeRepCapNameIds(instrSession, attributeId, bufferSize, repeatedCapabilityIds));

Error:
    return error;
}

/*****************************************************************************
* Function: Get Attribute Repeated Capability Id Name(s)
* Purpose:  This function returns the repeated capability name(s) that
*           belongs to the attribute and single repeated capability id you
*           may specify. More then one name is listed in comma separated
*           string.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_GetAttributeRepeatedCapabilityIdNames(
                            ViSession instrSession,
                            ViUInt32  attributeId,
                            ViString  repeatedCapabilityId,
                            ViInt32   bufferSize,
                            ViChar    repeatedCapabilityIdNames[])
{
    ViStatus error = VI_SUCCESS;

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, bufferSize, 1, RS_MAX_SHORT_MESSAGE_BUF_SIZE),
            4, "Buffer Size");

    checkErr(RsCore_GetAttributeRepCapNamesAll(instrSession,
        attributeId,
        repeatedCapabilityId,
        bufferSize,
        repeatedCapabilityIdNames));

Error:
    return error;
}

/*****************************************************************************
* Function: Reset
* Purpose:  This function resets the instrument.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_reset(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_ClearStatus(instrSession));
    checkErr(RsCore_Write(instrSession, "*RST")); // IGNORE: Check double write
    checkErr(RsCore_QueryViInt32(instrSession, "*OPC?", NULL));

    checkErr(rspwrmeter_DefaultInstrSetup(instrSession));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);

    return error;
}

/*****************************************************************************
* Function: Self Test
* Purpose:  This function executes the instrument self-test and returns the
*           result.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_self_test(ViSession instrSession,
                                       ViInt16*  testResult,
                                       ViChar    testMessage[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidNullPointer(instrSession, testResult), 2, "Null address for Test Result");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, testMessage), 3, "Null address for Test Message");

    checkErr(RsCore_SelfTest(instrSession, testResult, 48, testMessage));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: Revision Query
* Purpose:  This function returns the driver and instrument revisions.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_revision_query(ViSession instrSession,
                                            ViChar    instrumentDriverRevision[],
                                            ViChar    firmwareRevision[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidNullPointer(instrSession, instrumentDriverRevision), 2, "Null address for Driver Revision");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, firmwareRevision), 3, "Null address for Instrument Revision");

    checkErr(RsCore_GetAttributeViString(instrSession, NULL, RSPWRMETER_ATTR_SPECIFIC_DRIVER_REVISION,
        0, 256, instrumentDriverRevision));
    checkErr(RsCore_GetAttributeViString(instrSession, "", RSPWRMETER_ATTR_INSTRUMENT_FIRMWARE_REVISION,
        0, 256, firmwareRevision));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function:    System Version
* Purpose:     This function queries the number of the SCPI version, which
*              is relevant for the instrument.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_SystemVersion(ViSession instrSession,
                                           ViInt32   length,
                                           ViChar    systemVersion[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_VERSION,
            length, systemVersion),
            2, "System Version");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Sets a new security password.
///HINF
///HINF  The command is restricted to remote control over the USB
///HINF  interface (USBTMC).
///HIRET Returns the status code of this operation.
///HIPAR oldPassword/Old security password, entered as a string
///HIPAR newPassword/New security password, entered as a string
ViStatus _VI_FUNC rspwrmeter_SecurityPassword(ViSession instrSession,
                                              ViString  oldPassword,
                                              ViString  newPassword)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SYST:SEC:PASS:SEC '%s','%s'", oldPassword, newPassword);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Sets a new user password, also called instrument password.
///HINF
///HINF  The command is restricted to remote control over the USB
///HINF  interface (USBTMC).
///HIRET Returns the status code of this operation.
///HIPAR oldPassword/Old user password, entered as a string
///HIPAR newPassword/New user password, entered as a string
ViStatus _VI_FUNC rspwrmeter_UserPassword(ViSession instrSession,
                                          ViString  oldPassword,
                                          ViString  newPassword)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE] = "";

    checkErr(RsCore_LockSession(instrSession));

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SYST:SEC:PASS:USER '%s','%s'", oldPassword, newPassword);
    checkErr(RsCore_Write(instrSession, cmd));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Returns information about the system.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR length/Sets the allocated memory length of System Version string.
/// HIPAR systemInformation/Returns information about the system.
ViStatus _VI_FUNC rspwrmeter_SystemInformation(ViSession instrSession,
                                               ViInt32   length,
                                               ViChar    systemInformation[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_INFORMATION,
            length, systemInformation),
            2, "System Information");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets the system clock source.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR clockSource/Sets the system clock source.
ViStatus _VI_FUNC rspwrmeter_SystemClockSource(ViSession instrSession,
                                               ViInt32   clockSource)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_SYSTEM_CLOCK_SOURCE, clockSource),
            2, "Clock Source");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Configures sensor's LED.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR LEDMode/Selects whether the color of the system status LED is controlled by the firmware of the sensor internally or through the user settings.
/// HIPAR red/Red component.
/// HIPAR green/Green component.
/// HIPAR blue/Blue component.
ViStatus _VI_FUNC rspwrmeter_ConfigureLED(ViSession instrSession,
                                          ViInt32   LEDMode,
                                          ViInt32   red,
                                          ViInt32   green,
                                          ViInt32   blue)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, red, 0, 255),
            3, "Red");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, green, 0, 255),
            4, "Green");
    viCheckParm(RsCore_InvalidViInt32Range(instrSession, blue, 0, 255),
            5, "Blue");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_LED_MODE, LEDMode),
            2, "LED Mode");

    if (LEDMode == RSPWRMETER_VAL_LED_USER)
    {
        ViUInt32 color = blue;

        color |= (red << 16) | (green << 8);

        viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_LED_COLOR, color),
                3, "LED Color");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Yields the lower power measurement limit. This value may change if [SENSe<Sensor>:]CORRection:SPDevice:STATe is set to ON. The lower measurement limit may refer to the sensor or to the combination of a sensor and the components connected ahead of it.
/// HIFN This query can be used to determine a useful resolution for the result display near the lower measurement limit.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR lowerPowerMeasurementLimit/Yields the lower power measurement limit. This value may change if [SENSe<Sensor>:]CORRection:SPDevice:STATe is set to ON. The lower measurement limit may refer to the sensor or to the combination of a sensor and the components connected ahead of it. This query can be used to determine a useful resolution for the result display near the lower measurement limit.
ViStatus _VI_FUNC rspwrmeter_QueryLowerPowerMeasurementLimit(ViSession instrSession,
                                                             ViReal64* lowerPowerMeasurementLimit)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViReal64(instrSession, "", RSPWRMETER_ATTR_SYSTEM_LOWER_POWER_MEASUREMENT_LIMIT, lowerPowerMeasurementLimit),
            2, "Lower Power Measurement Limit");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the IP address settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR mode/Configures the system network IP address mode.
/// HIPAR ipAddress/Configures the system network IP address.
/// HIPAR subnetMask/Configures the system network IP address subnet mask.
/// HIPAR gateway/Configures the system network IP address gateway.
ViStatus _VI_FUNC rspwrmeter_ConfigureIPAddress(ViSession instrSession,
                                                ViInt32   mode,
                                                ViString  ipAddress,
                                                ViString  subnetMask,
                                                ViString  gateway)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_IP_ADDRESS_MODE, mode),
            2, "Mode");

    if (mode == RSPWRMETER_VAL_IP_ADDRESS_STATIC)
    {
        viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_IP_ADDRESS, ipAddress),
                3, "IP Address");

        viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_IP_ADDRESS_SUBNET_MASK, subnetMask),
                4, "Subnet Mask");

        viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_IP_ADDRESS_GATEWAY, gateway),
                5, "Gateway");
    }

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the system network IP address info.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR length/Sets the allocated memory length of Info string.
/// HIPAR info/Returns the system network IP address info.
ViStatus _VI_FUNC rspwrmeter_GetIPAddressInfo(ViSession instrSession,
                                              ViInt32   length,
                                              ViChar    info[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_IP_ADDRESS_INFO,
            length, info),
            3, "Info");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Returns the system network status.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR length/Sets the allocated memory length of Status string.
/// HIPAR status/Returns the system network status.
ViStatus _VI_FUNC rspwrmeter_GetNetworkStatus(ViSession instrSession,
                                              ViInt32   length,
                                              ViChar    status[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_STATUS,
            length, status),
            3, "Status");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Resets the system network settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_ResetNetworkSettings(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_RESET, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Restarts system network.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
ViStatus _VI_FUNC rspwrmeter_RestartNetworkSettings(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_RESTART, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the common settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR domain/Configures the system network domain.
/// HIPAR hostname/Configures the system network hostname.
ViStatus _VI_FUNC rspwrmeter_ConfigureCommonSettings(ViSession instrSession,
                                                     ViString  domain,
                                                     ViString  hostname)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_DOMAIN, domain),
            2, "Domain");

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_HOSTNAME, hostname),
            3, "Hostname");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function configures the socket settings.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR mode/Configures system socket mode.
/// HIPAR port/Configures system socket port.
ViStatus _VI_FUNC rspwrmeter_ConfigureSocket(ViSession instrSession,
                                             ViInt32   mode,
                                             ViInt32   port)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_SOCKET_MODE, mode),
            2, "Mode");

    viCheckParm(rspwrmeter_SetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_SYSTEM_NETWORK_SOCKET_PORT, port),
            3, "Port");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}



/*****************************************************************************
* Function: Error Query
* Purpose:  This function queries the instrument error queue and returns
*           the result.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_error_query(ViSession instrSession,
                                         ViInt32*  errorCode,
                                         ViChar    errorMessage[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidNullPointer(instrSession, errorCode), 2, "Null address for Error Code");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

    checkErr(RsCore_ErrorQueryAll(instrSession, errorCode, 1024, errorMessage));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: Error Message
* Purpose:  This function translates the error codes returned by this
*           instrument driver into user-readable strings.
*
*           Note:  The caller can pass NULL for the vi parameter.  This
*           is useful if one of the init functions fail.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_error_message(ViSession instrSession,
                                           ViStatus  errorCode,
                                           ViChar    errorMessage[])
{
    return rspwrmeter_utility_error_message(instrSession,errorCode,errorMessage);
}

/*****************************************************************************
* Function: Preset
* Purpose:  This function initiates an instrument reset.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_Preset(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_SetAttributeViString(instrSession, "", RSPWRMETER_ATTR_PRESET_DEVICE, NULL));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Queries the number of entries in the error queue.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR numberOfErrors/Returns the number of entries in the error queue.
ViStatus _VI_FUNC rspwrmeter_ReadErrorCount(ViSession instrSession,
                                            ViInt32*  numberOfErrors)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_GetAttributeViInt32(instrSession, "", RSPWRMETER_ATTR_SYSTEM_ERROR_COUNT, numberOfErrors));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Queries all unread entries in the error/event queue and removes them from the queue. The response is a comma separated list of error numbers and a short description of the error in the first
/// HIFN in first out order. Positive error numbers are instrument-dependent. Negative error numbers are reserved by the SCPI standard.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR numberOfBytesToRead/Pass the maximum number of bytes to read from the instruments.
/// HIPAR readBuffer/After this function executes, this parameter contains the data that was read from the instrument.
/// HIPAR numBytesRead/Returns the number of bytes actually read from the instrument and stored in the Read Buffer.
ViStatus _VI_FUNC rspwrmeter_ReadAllErrors(ViSession instrSession,
                                           ViUInt32  numberOfBytesToRead,
                                           ViChar    readBuffer[],
                                           ViUInt32* numBytesRead)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(rspwrmeter_QueryViString(instrSession, "SYST:ERR:ALL?", numberOfBytesToRead, readBuffer));

    if (numBytesRead)
        *numBytesRead = (ViInt32)strlen(readBuffer);

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Returns the list of static error (if any). Static errors are generally more severe than normal error conditions, which can be queried with SYSTem:ERRor[:NEXT]?.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
ViStatus _VI_FUNC rspwrmeter_ReadAllStaticErrors(ViSession instrSession,
                                                 ViInt32*  errorCode,
                                                 ViChar    errorMessage[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidNullPointer(instrSession, errorMessage), 3, "Null address for Error Message");

    if (RsCore_Simulating(instrSession))
    {
        strcpy(errorMessage, "No error.");
        *errorCode = 0;
        goto Error;
    }

    checkErr(rspwrmeter_QueryViString(instrSession, "SYST:SERR?", 1024, errorMessage));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN This function sets the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR timeout/This control sets the timeout value.
ViStatus _VI_FUNC rspwrmeter_SetTimeout(ViSession instrSession,
                                        ViUInt32  timeout)
{  
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_SetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, (ViInt32)timeout));

Error:
    return error;
}


/// HIFN This function returns the OPC timeout.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.
/// HIPAR timeout/This control returns the timeout value.
ViStatus _VI_FUNC rspwrmeter_GetTimeout(ViSession instrSession,
                                        ViUInt32* timeout)
{  
    ViStatus error = VI_SUCCESS;
    ViInt32  timeoutInt;

    checkErr(RsCore_GetAttributeViInt32(instrSession, NULL, RS_ATTR_OPC_TIMEOUT, 0, &timeoutInt));
    *timeout = (ViUInt32)timeoutInt;

Error:
    return error;
}

/// HIFN This function specifies the minimum timeout value to use (in
/// HIFN milliseconds) when accessing the device associated with the
/// HIFN given session.
/// HIFN Note that the actual timeout used may be higher than the one
/// HIFN requested.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR VISATimeout/Specifies the minimum timeout value to use (in milliseconds)
/// HIPAR VISATimeout/when accessing the device associated with the given session.
/// HIPAR VISATimeout/Note that the actual timeout used may be higher than the one
/// HIPAR VISATimeout/requested.
ViStatus _VI_FUNC rspwrmeter_SetVISATimeout(ViSession instrSession,
                                            ViUInt32  VISATimeout)
{
    return RsCore_SetVisaTimeout(instrSession, VISATimeout);
}

/// HIFN This function queries the minimum timeout value to use (in
/// HIFN milliseconds) when accessing the device associated with the
/// HIFN given session.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function. The handle identifies a
/// HIPAR instrSession/particular instrument session.
/// HIPAR VISATimeout/Queries the minimum timeout value to use (in milliseconds)
/// HIPAR VISATimeout/when accessing the device associated with the given session.
ViStatus _VI_FUNC rspwrmeter_GetVISATimeout(ViSession instrSession,
                                            ViUInt32* VISATimeout)
{
    return RsCore_GetVisaTimeout(instrSession, VISATimeout);
}

/// HIFN This function switches ON/OFF various error checkings performed inside of the driver.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/This control accepts the Instrument Handle returned by the Initialize
/// HIPAR instrSession/function to select the desired instrument driver session.
/// HIPAR optionChecking/This control switches option checking On or Off.
/// HIPAR rangeChecking/This control enables or disables range checking.
/// HIPAR statusChecking/This control enables or disables instrument state checking
ViStatus _VI_FUNC rspwrmeter_ConfigureErrorChecking(ViSession instrSession,
                                                    ViBoolean optionChecking,
                                                    ViBoolean rangeChecking,
                                                    ViBoolean statusChecking)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession = NULL;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));

    rsSession->optionChecking = optionChecking;
    checkErr(RsCore_SetAttributeViBoolean(instrSession, "", RS_ATTR_QUERY_INSTRUMENT_STATUS, 0, statusChecking));
    checkErr(RsCore_SetAttributeViBoolean(instrSession, "", RS_ATTR_RANGE_CHECK, 0, rangeChecking));

Error:
    return error;
}

/// HIFN  This function clears status.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rspwrmeter_ClearStatus(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_ClearStatus(instrSession));
    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function returns the ID Query response string.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR bufferSize/Size of IDQueryResponse buffer
/// HIPAR IDQueryResponse/Returns the ID Query response string. The array should consist of at
/// HIPAR IDQueryResponse/least 256 elements.
ViStatus _VI_FUNC rspwrmeter_IDQueryResponse(ViSession instrSession,
                                             ViUInt32  bufferSize,
                                             ViChar    IDQueryResponse[])
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    if (RsCore_Simulating(instrSession))
    {
        strncpy(IDQueryResponse, RSPWRMETER_SIMULATION_ID_QUERY, bufferSize);
        IDQueryResponse[bufferSize - 1] = 0;
        goto Error;
    }

    checkErr(rspwrmeter_QueryViString(instrSession, "*IDN?", bufferSize, IDQueryResponse));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Stops further command processing until all commands sent before *WAI
/// HIFN  have been executed.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
ViStatus _VI_FUNC rspwrmeter_ProcessAllPreviousCommands(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_Write(instrSession, "*WAI"));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  This function queries the OPC.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.  The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR opc/Queries the OPC.
ViStatus _VI_FUNC rspwrmeter_QueryOPC(ViSession instrSession,
                                      ViInt32*  opc)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_QueryViInt32(instrSession, "*OPC?", opc));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Reads the footprint file of the sensor.
/// HIFN
/// HIFN     Remote-control command(s):
/// HIFN     SYSTem:DFPRint<Channel>?
/// HIFN
/// HIFN     Attribute(s):
/// HIFN     RSPWRMETER_FUNC_DEVICE_FOOTPRINT
/// HIFN
/// HIFN     Remote-control command(s):
/// HIFN
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel/Measurement channel if more than one channel is available.
/// HIPAR bufferSize/Defines the size of buffer in argument 'Device Footprint'.
/// HIPAR deviceFootprint/
/// HIPAR returnedValues/Returns count of characters read from instrument and stored in buffer
/// HIPAR returnedValues/passed to 'Measurement'.
ViStatus _VI_FUNC rspwrmeter_GetDeviceFootprint(ViSession instrSession,
                                                ViInt32   channel,
                                                ViUInt32  bufferSize,
                                                ViChar    deviceFootprint[],
                                                ViUInt32* returnedValues)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];
    ViByte*  data = NULL;
    ViInt64  count = 0;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, channel, 0, 4),
            2, "Channel");
    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SYST:DFPR%ld?", channel);
    checkErr(RsCore_QueryBinaryDataBlock(instrSession, cmd, VI_TRUE, NULL, &data, &count));

    if (returnedValues != NULL)
        *returnedValues = (ViUInt32)count;

    checkErr(rspwrmeter_CheckStatus(instrSession));

    checkErr(RsCore_CopyToUserBufferBinData((ViBuf)deviceFootprint, bufferSize, data, count));

Error:
    if (data)
        free(data);

    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Queries the possible power test levels of the sensor.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR systemTestLevels/Queries the possible power test levels of the sensor.
ViStatus _VI_FUNC rspwrmeter_GetSystemTestLevels(ViSession instrSession,
                                                 ViInt32*  systemTestLevels)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_GetAttributeViInt32(instrSession, NULL, RSPWRMETER_ATTR_SYSTEM_TEST_LEVELS, systemTestLevels),
            2, "System Test Levels");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN  Sets a name of the sensors according to the users requirements. The
/// HIFN  name specified by 'new-name' will be displayed on the WebGUI of the
/// HIFN  network sensors.
/// HIFN     The sensor name given here is independent from the hostname of the
/// HIFN  sensor. However, if the sensor name is not yet specified it defaults to
/// HIFN  the hostname.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR systemSensorName/Sets a name of the sensors according to the users requirements. The
/// HIPAR systemSensorName/name specified by 'new-name' will be displayed on the WebGUI of the
/// HIPAR systemSensorName/network sensors.
ViStatus _VI_FUNC rspwrmeter_SetSystemSensorName(ViSession instrSession,
                                                 ViString  systemSensorName)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(rspwrmeter_SetAttributeViString(instrSession, NULL, RSPWRMETER_ATTR_SYSTEM_SENSOR_NAME, systemSensorName),
            2, "System Sensor Name");

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN If TRUE (default after init is FALSE), the driver calls SYST:ERR? automatically in CheckStatus() when Instrument Status Error is detected.
/// HIFN Use the function rspwrmeter_GetError afterwards to get the detailed information about the instrument error(s).
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR autoSystErrQuery/If TRUE (default after init is FALSE), the driver calls SYST:ERR? automatically in CheckStatus() when Instrument Status Error is detected.
ViStatus _VI_FUNC rspwrmeter_ConfigureAutoSystemErrQuery(ViSession instrSession,
                                                         ViBoolean autoSystErrQuery)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession = NULL;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    rsSession->autoSystErrQuery = autoSystErrQuery;

Error:
    return error;
}

/// HIFN Returns TRUE, if the current instrument is a Base Unit NRP2 or NRX
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR isBaseUnit/Returns TRUE, if the current instrument is a Base Unit NRP2 or NRX
ViStatus _VI_FUNC rspwrmeter_IsBaseUnit(ViSession  instrSession,
                                        ViBoolean* isBaseUnit)
{  
    ViStatus error = VI_SUCCESS;

    *isBaseUnit = RsCore_IsInstrumentModel(instrSession, "NRX|NRP2|NRP");

    return error;
}

/// HIFN For a base unit returns TRUE if a channel has a sensor attached.
/// HIFN For a single sensor returns TRUE for channel1
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR channel1/Returns TRUE if the sensor is present on base unit's channel 1
/// HIPAR channel2/Returns TRUE if the sensor is present on base unit's channel 2
/// HIPAR channel3/Returns TRUE if the sensor is present on base unit's channel 3
/// HIPAR channel4/Returns TRUE if the sensor is present on base unit's channel 4
ViStatus _VI_FUNC rspwrmeter_GetBaseUnitSensorsPresence(ViSession instrSession,
                                                        ViBoolean *channel1,
                                                        ViBoolean *channel2,
                                                        ViBoolean *channel3,
                                                        ViBoolean *channel4)
{
    ViStatus  error = VI_SUCCESS;
    ViChar    response[RS_MAX_SHORT_MESSAGE_BUF_SIZE];
    ViInt32   registerValue = 0;
    ViBoolean isBaseUnit;

    (void)RsCore_LockSession(instrSession);

    *channel1 = VI_FALSE;
    *channel2 = VI_FALSE;
    *channel3 = VI_FALSE;
    *channel4 = VI_FALSE;

    checkErr(rspwrmeter_IsBaseUnit(instrSession, &isBaseUnit));
    if (isBaseUnit == VI_FALSE)
    {
        *channel1 = VI_TRUE;
        return error;
    }

    checkErr(RsCore_QueryViStringShort(instrSession, "STAT:DEV:COND?", response));
    registerValue = atoi(response);

    if ((registerValue & 2) > 0) *channel1 = VI_TRUE;
    if ((registerValue & 4) > 0) *channel2 = VI_TRUE;
    if ((registerValue & 128) > 0) *channel3 = VI_TRUE;
    if ((registerValue & 256) > 0) *channel4 = VI_TRUE;

Error:  
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/// HIFN Session-related mutex switch.
/// HIFN If TRUE (default after init is FALSE), the driver sets a mutual-exclusion mechanism for every driver's hi-level function and attribute access.
/// HIFN Set this value to TRUE if you plan to use one session in more than one thread.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR multiThreadLocking/Session-related mutex switch
ViStatus _VI_FUNC rspwrmeter_ConfigureMultiThreadLocking(ViSession instrSession,
                                                         ViBoolean multiThreadLocking)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession = NULL;

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    rsSession->multiThreadLocking = multiThreadLocking;

Error:
    return error;
}

/// HIFN  Saves all commands, or those that differ from the defined default
/// HIFN  status set by *RST, to an xml file.
/// HIRET Returns the status code of this operation.
/// HIPAR instrSession/The ViSession handle that you obtain from the rspwrmeter_init or
/// HIPAR instrSession/rspwrmeter_InitWithOptions function.The handle identifies a particular
/// HIPAR instrSession/instrument session.
/// HIPAR type/Selects the format of commands to be returned
/// HIPAR targetPCFile/Specifies a local file to save the specified data to.
ViStatus _VI_FUNC rspwrmeter_SaveCommands(ViSession instrSession,
                                          ViInt32   type,
                                          ViString  targetPCFile)
{
    ViStatus error = VI_SUCCESS;
    ViChar   cmd[RS_MAX_MESSAGE_BUF_SIZE];

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, type, RSPWRMETER_VAL_ALL_CMD, RSPWRMETER_VAL_DIFFERENT_CMD),
            2, "Type");

    snprintf(cmd, RS_MAX_MESSAGE_BUF_SIZE, "SYST:PAR%s?", (type == RSPWRMETER_VAL_ALL_CMD) ? "" : ":DELT");
    checkErr(RsCore_QueryBinaryDataBlockToFile(instrSession, cmd, targetPCFile, RS_VAL_TRUNCATE));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Switches OPC-synchronization ON/OFF for writes.
///HINF
///HIRET Returns the status code of this operation.
///HIPAR enable/Switches OPC-synchronization ON/OFF for writes.
ViStatus _VI_FUNC rspwrmeter_OPCSyncWriteEnable(ViSession instrSession,
                                                ViBoolean enable)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    rsSession->opcSyncAttrWrites = enable;

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Switches OPC-synchronization ON/OFF for queries.
///HINF
///HIRET Returns the status code of this operation.
///HIPAR enable/Switches OPC-synchronization ON/OFF for queries.
ViStatus _VI_FUNC rspwrmeter_OPCSyncQueryEnable(ViSession instrSession,
                                                ViBoolean enable)
{
    ViStatus         error = VI_SUCCESS;
    RsCoreSessionPtr rsSession;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_GetRsSession(instrSession, &rsSession));
    rsSession->opcSyncAttrQueries = enable;

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

///HINF  Sets the mechanism which is used for OPC-synchronised queries
///HINF  (valid for read attributes and all OPC-synchronised queries):
///HINF  Default value: 'useEsb' = true, 'clsBefore' = true, 'useMav' =
///HINF  false
///HINF
///HINF  Scenario A: useEsb = true:
///HINF  The query command is appended by the ';*OPC', and the wait
///HINF  loop ends if the ESB bit signals that the command was completed,
///HINF  or the ERRQ (bit 2) signals an error.
///HINF      - clsBefore has no effect and is always set to true.
///HINF      - set the 'useMav' to true, to enable the opc-wait loop
///HINF  sucessfully finish if the MAV bit signals, that the response is
///HINF  available, even if the ESB bit is still 0.
///HINF
///HINF  Scenario B: useEsb = false:
///HINF  The query command is not appended by the ';*OPC', and the wait
///HINF  loop ends if the MAV bit signals that the message is
///HINF  available, or the ERRQ (bit 2) signals an error.
///HINF      - clsBefore causes sending query *STB? before sending the
///HINF  command to clear the error queue and output buffers.
///HINF      - useMav has no effect, and is always set to true.
///HINF
///HIRET Returns the status code of this operation.
///HIPAR useEsb/Switches queries OPC-synchronization to use the Event
///HIPAR useEsb/Status Bit.
///HIPAR clsBefore/Switches queries OPC-synchronization to use call
///HIPAR clsBefore/the Clear Before Read procedure before sending the
///HIPAR clsBefore/query.
///HIPAR useMav/Switches queries OPC-synchronization to react on
///HIPAR useMav/Message Available bit (MAV)
ViStatus _VI_FUNC rspwrmeter_SetOpcSyncQueryMechanism(ViSession instrSession,
                                                      ViBoolean useEsb,
                                                      ViBoolean clsBefore,
                                                      ViBoolean useMav)
{
    return RsCore_SetOpcSyncQueryMechanism(instrSession, useEsb, clsBefore, useMav);
}

/*****************************************************************************
* Function: Get Error and Clear Error Functions
* Purpose:  These functions enable the instrument driver user to
*           get or clear the error information the driver associates with the
*           RS session.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_GetError(ViSession instrSession,
                                      ViStatus* errorCode,
                                      ViInt32   bufferSize,
                                      ViChar    description[])
{
    ViStatus error = VI_SUCCESS;

    // Lock Session - do not jump to the end, even with invalid instrSession the function continues further
    (void)(RsCore_LockSession(instrSession));

    // Test for nulls and acquire error data
    viCheckParm(RsCore_InvalidNullPointer(instrSession, errorCode), 2, "Null address for Error");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, description), 4, "Null address for Description");

    checkErr(RsCore_GetErrorCompleteDescription(instrSession, &rspwrmeter_utility_error_message, errorCode, bufferSize, description));

Error:
    // Unlock Session
    (void)RsCore_UnlockSession(instrSession); // IGNORE: ERROR!!! Missing Lock
    return error;
}

ViStatus _VI_FUNC rspwrmeter_ClearError(ViSession instrSession)
{
    return RsCore_ClearErrorInfo(instrSession);
}

/*****************************************************************************
* Function: WriteInstrData and ReadInstrData Functions
* Purpose:  These functions enable the instrument driver user to
*           write and read commands directly to and from the instrument.
*
*           Note:  These functions bypass the RS attribute state caching.
*                  WriteInstrData invalidates the cached values for all
*                  attributes.
*****************************************************************************/

ViStatus _VI_FUNC rspwrmeter_viWrite(ViSession instrSession,
                                     ViString  writeBuffer,
                                     ViUInt32  count,
                                     ViUInt32* returnCount)
{
    ViStatus error = VI_SUCCESS;
             count;

    checkErr(RsCore_LockSession(instrSession));

    checkErr(RsCore_Write(instrSession, writeBuffer));

    if (returnCount)
        *returnCount = (ViUInt32)strlen(writeBuffer);

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: Query ViBoolean
* Purpose:  This function queries the ViBoolean value.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_QueryViBoolean(ViSession  instrSession,
                                            ViString   command,
                                            ViBoolean* value)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(command), 1, 10000000), 2, "Command(null string length)");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

    checkErr(RsCore_QueryViBoolean(instrSession, command, value));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: Query ViInt32
* Purpose:  This function queries the ViInt32 value.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_QueryViInt32(ViSession instrSession,
                                          ViString  command,
                                          ViInt32*  value)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(command), 1, 10000000), 2, "Command(null string length)");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

    checkErr(RsCore_QueryViInt32(instrSession, command, value));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: Query ViReal64
* Purpose:  This function queries the ViReal64 value.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_QueryViReal64(ViSession instrSession,
                                           ViString  command,
                                           ViReal64* value)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(command), 1, 10000000), 2, "Command(null string length)");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

    checkErr(RsCore_QueryViReal64(instrSession, command, value));

    checkErr(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}


/*****************************************************************************
* Function: Query ViString
* Purpose:  This function queries the ViString value.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_QueryViString(ViSession instrSession,
                                           ViString  command,
                                           ViInt32   bufferSize,
                                           ViChar    value[])
{
    ViStatus error = VI_SUCCESS;
    ViStatus warning = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));

    viCheckParm(RsCore_InvalidViInt32Range(instrSession, (ViInt32)strlen(command), 1, 10000000), 2, "Command(null string length)");
    viCheckParm(RsCore_InvalidNullPointer(instrSession, value), 3, "Value");

    checkErr(RsCore_QueryViStringUnknownLengthToUserBuffer(instrSession, command, bufferSize, value, NULL));
    checkErrKeepWarn(rspwrmeter_CheckStatus(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    return error;
}

/*****************************************************************************
* Function: rspwrmeter_close
* Purpose:  This function closes the instrument.
*
*           Note:  This function must unlock the session before calling
*           Rs_Dispose.
*****************************************************************************/
ViStatus _VI_FUNC rspwrmeter_close(ViSession instrSession)
{
    ViStatus error = VI_SUCCESS;

    checkErr(RsCore_LockSession(instrSession));  
    checkErr(RsCore_ViClose(instrSession));

Error:
    (void)RsCore_UnlockSession(instrSession);
    (void)RsCore_Dispose(instrSession);

    return error;
}

